from ..default import *
from .. import structure
from .basic import * 
from ..statistics import stats_methods

import scipy.fftpack
from scipy import signal

comparison_params = dict(test=stats_methods.t_test, win='1ms', method='mean', sig_limit=0.05, need_fdr=False)

def Spectrum(self, compare=False, freq_span=(0,30), target='power', comparison_params=comparison_params):
    # with the decorator, we can just focuse on case data instead of batch/collection data
    @self.iter('average')
    def to_spetrum(case_raw_data):
        erp = case_raw_data.mean(level=['subject','condition_group','channel_group'])

        Spectrum_df = []
        for name,data in erp.groupby(level=['subject','condition_group','channel_group']):
            N = data.shape[1]

            fft_result = scipy.fftpack.fft(data,axis=1)
            if target=='power':
                fft_result = 2.0/N * np.abs(fft_result)
            elif target=='phase':
                fft_result = np.angle(fft_result)
            else:
                raise Exception('Please set the parameter "target" as "power", or "phase".')
            fft_result = fft_result[:,:N//2]
            if freq_span[0]==0:
                fft_result[:,0] = 0

            index = pd.MultiIndex.from_tuples([name], 
                names=['subject','condition_group','channel_group'])

            fft_result = pd.DataFrame(fft_result,
                index=index,
                columns=np.linspace(0, self.info['sample_rate'], N//2)) # resolution: sr/N
            fft_result = fft_result.loc[:, freq_span[0]:freq_span[1]] 
            
            fft_result.columns = pd.MultiIndex.from_tuples([(0, freq) for freq in fft_result.columns], names=['','frequency'])
            Spectrum_df.append(fft_result)
        
        Spectrum_df = pd.concat(Spectrum_df)
        return Spectrum_df
    
    spetrum_collection = to_spetrum()
    if compare:
        stats_data = [stats_compare(spetrum_batch, comparison_params, levels='frequency', between='condition_group', in_group='subject') 
                        for spetrum_batch in spetrum_collection]
    else:
        stats_data = None

    default_plot_params = dict(plot_type=['direct','spectrum'], y_title='Spectrum', err_style='ci_band', color="Set1", style='darkgrid',win=comparison_params['win'],sig_limit=0.05)
    return structure.Analyzed_data('Spectrum', spetrum_collection, stats_data, default_plot_params)

# grand average
def Time_frequency(self, compare=False, freq_span=(0, 30)):
    # with the decorator, we can just focuse on case data instead of batch/collection data
    @self.iter('average')
    def to_tf(case_raw_data):
        erp = case_raw_data.mean(level=['condition_group','channel_group'])

        tf_df = []
        for name,data in erp.groupby(level=['condition_group','channel_group']):
            freqs = np.arange(freq_span[0], freq_span[1])
            if freq_span[0]==0:
                widths = freqs+0.001
            else:
                widths = freqs
            # index = pd.MultiIndex.from_tuples([(*name,freq) for freq in freqs[::-1]], 
            #     names=['condition_group','channel_group','frequency'])
    
            cwt_result = signal.cwt(np.array(data)[0], signal.ricker, widths=widths)
            cwt_result = pd.DataFrame(cwt_result, columns=data.columns)
            cwt_result.index = pd.MultiIndex.from_tuples(
                [(name[0], name[1], i) for i in freqs[::-1]], names=('condition_group', 'channel_group', 'freq'))
            
            tf_df.append(cwt_result)
        
        tf_df = pd.concat(tf_df)

        for level in ['condition_group', 'channel_group']:
            if len(tf_df.index.get_level_values(level).unique()) == 2:
                tf_df = [i for ind,i in tf_df.groupby(level=level)]
                tf_df[0].index = tf_df[0].index.droplevel(level)
                tf_df[1].index = tf_df[1].index.droplevel(level)
                tf_df = tf_df[0] - tf_df[1]
        
        tf_df.index = tf_df.index.get_level_values('freq')

        return tf_df

    @self.iter('average')
    def to_tf_group_level(case_raw_data):
        erp = case_raw_data.mean(level=['condition_group', 'channel_group','subject'])

        tf_df = []
        for name, data in erp.groupby(level=['condition_group', 'channel_group', 'subject']):
            
            freqs = np.arange(freq_span[0], freq_span[1])
            if freq_span[0] == 0:
                widths = freqs+0.001
            else:
                widths = freqs

            cwt_result = signal.cwt(
                np.array(data)[0], signal.ricker, widths=widths)
            
            cwt_result = pd.DataFrame(
                cwt_result, columns=data.columns)
            cwt_result.index = pd.MultiIndex.from_tuples(
                [(name[0], name[1], name[2], i) for i in freqs[::-1]], names=('condition_group', 'channel_group', 'subject', 'freq'))
            
            tf_df.append(cwt_result)

        tf_df = pd.concat(tf_df)
        return tf_df
    
    tf_collection = to_tf()

    if compare:
        tf_group_level_collection = to_tf_group_level()
        # print(tf_group_level_collection)
        stats_data = [stats_compare(tf_batch, comparison_params, levels=['time', 'freq'], between='condition_group', in_group='subject')
                      for tf_batch in tf_group_level_collection]
    else:
        stats_data = None

    default_plot_params = dict(plot_type=['direct', 'heatmap'], x_title='time', y_title='frequency', sig_limit=0.05,
        color="RdBu_r", style='white', grid=False, cbar_title='Power')
    return structure.Analyzed_data('Time Frequency', tf_collection, stats_data, default_plot_params)



from ..default import *

def multiple_comparison_correction(pvs_list, method='fdr_bh'):
    if method == 'smooth':
        def correct(v):
            level = 0.05
            v = np.concatenate(([1],v,[1]))
            for idx in range(1,len(v)-1):
                if v[idx] < level and v[idx-1] >= level and v[idx+1] >= level:
                    v[idx] = level
            return v[1:-1]
    else:
        correct = lambda v: statsmodels.sandbox.stats.multicomp.multipletests(v, 0.05, method)[1]
    
    pvs_list_new = []
    for pvs in pvs_list:
        pvs_list_new.append(pd.DataFrame([correct(i) for i in pvs.values], index=pvs.index, columns=pvs.columns))
        pvs_list_new[-1].name = pvs.name
        
    return pvs_list_new

def t_test(values):
    a,b = values
    t, pv = scipy.stats.ttest_rel(a,b)
    return {'pvalue':pv, 'effect':t}

def permutation_test(values,reps=1000):
    values = np.array(values)
    if len(values)==2:
        pv,t = two_sample(values[0],values[1], reps=reps,stat=lambda u,v: mean(u-v),alternative='two-sided')
        return {'pvalue':pv, 'effect':t}
    else:
        return None

def get_pvalue_from_distribution(result_real, dist_baseline):
    dist_baseline.append(result_real)
    dist_baseline.sort()
    
    pvalue = 1-dist_baseline.index(result_real)/len(dist_baseline)
    if pvalue == 0:
        pvalue += 1/len(dist_baseline)

    return pvalue

def permutation_on_condition(data,method,shuffle_count=1000):
    def condition_shuffled(data):
        group_labels = list(data.index.get_level_values(level='cond_group'))
        random.shuffle(group_labels)
        data.index = data.index.set_labels(group_labels,level='cond_group')
        return data
    # keep
    val_raw = method(data)

    # shuffle
    baseline =[]
    for i in range(shuffle_count):
        val_shuffled = method(condition_shuffled(data))
        baseline.append(val_shuffled)

    baseline.append(2) # correction
    baseline.append(val_raw)
    baseline.sort()

    pv = 1-baseline.index(val_raw)/shuffle_count

    return pv,None

def fdr(pvs):
    re_calc = lambda v: statsmodels.sandbox.stats.multicomp.multipletests(v, 0.05, 'fdr_bh')[1]
    if type(pvs) is pd.DataFrame:
        fdr = [re_calc(i) for i in pvs.values]
        pvs_new = pd.DataFrame(fdr, index=pvs.index, columns=pvs.columns)
    elif type(pvs) is pd.Series:
        fdr = re_calc(pvs.values)
        pvs_new = pd.Series(fdr, index=pvs.index, name=pvs.name)
    elif type(pvs) is dict:
        fdr = re_calc(pvs.values())
        pvs_new = dict(zip(pvs.keys(),fdr))
    elif type(pvs) is list:
        pvs_new = re_calc(pvs)
    else:
        raise ValueError('Only support pd.DataFrame, pd.Series,dict, and list')
    return pvs_new

def anova():
    pass

def similarity():
    pass

def classifier():
    pass            


from .default import *
from . import io
from . import group

class Epochs():
    def __init__(self, epochs_data, montage_path='standard-10-5-cap385', info=dict()):
        if type(epochs_data) != pd.DataFrame:
            raise Exception('Unsupported input!')

        if 'sample_rate' not in info:
            # inferred_freq = epochs_data.columns.inferred_freq
            # if len(inferred_freq)==1:
            #     inferred_freq = '1'+inferred_freq # sometimes inferred_freq is "L"
            # if inferred_freq[-1] == 'L':
            #     sr = int(1000/float(inferred_freq[:-1]))
            # elif inferred_freq[-1] == 'U':
            #     sr = int(1000000/float(inferred_freq[:-1]))
            info['sample_rate'] = 1000//(epochs_data.columns[1]-epochs_data.columns[0])

        info['subjects'] = {'all':list(epochs_data.index.get_level_values('subject').unique())}
        info['timepoints'] = {'all':list(epochs_data.columns)}

        info['conditions'] = dict()
        info['conditions']['all'] = list(epochs_data.index.get_level_values('condition').unique())

        info['channels'] = dict()
        info['channels']['all'] = list(epochs_data.index.get_level_values('channel').unique())
        
        info['trials'] = dict()
        info['trials']['all'] = list(epochs_data.index.get_level_values('trial').unique())

        for subj_id,subj_data in epochs_data.groupby(level=['subject']):
            info['conditions'][str(subj_id)] = sorted(list(subj_data.index.get_level_values('condition').unique()))
            info['channels'][str(subj_id)] = sorted(list(subj_data.index.get_level_values('channel').unique()))
            info['trials'][str(subj_id)] = sorted(list(subj_data.index.get_level_values('trial').unique()))

        if 'channel&id' not in info:
            info['channel&id'] = dict() # build up a dictionary contained id2channel and channel2id
            for ind,i in enumerate(info['channels']['all']):
                info['channel&id'][str(i)] = ind
                info['channel&id'][ind] = str(i)
           
        if 'xy_locs' not in info:
            info['xy_locs'] = io.load_topolocs(montage_path, info['channels']['all'])

        epochs_data_averaged = epochs_data.mean(level=list(np.setdiff1d(epochs_data.index.names, ['trial'])))
        epochs_data_averaged.index = pd.MultiIndex.from_tuples([(subject,condition,info['conditions']['all'].index(condition),channel) for (channel,condition,subject) in epochs_data_averaged.index],
                                 names=['subject','condition','trial','channel'])
        epochs_data_averaged.sort_index(inplace=True)

        self.all = epochs_data
        self.average = epochs_data_averaged
        self.info = info

    def extract(self, collection_script):
        collection_frame = group.parsing(collection_script, self)    
        return Extracted_epochs(self, collection_frame)

    def save(self, filepath, append=False):
        io.save_epochs(self, filepath, append=False, all_in_one=False)

class Extracted_epochs():
    def __init__(self, epochs, collection_frame):
        self.data = epochs
        self.frame = collection_frame
        self.info = epochs.info

    def iter_batchs(self):
        for batch_name, batch_frame in self.frame:
            yield batch_name, batch_frame

    def iter_cases(self, batch_frame):
        for case_frame in batch_frame:
            yield case_frame

    def get_batch_names(self, batch_id='all'):
        batch_names = [batch_name for batch_name,batch_frame in self.frame]

        if batch_id == 'all':
            return batch_names
        elif type(batch_id) is list:
            return [batch_names[idx] for idx in batch_id]
        elif type(batch_id) is int:
            return batch_names[batch_id]

    def get_case_name(self, batch_id=0, case_id=0):
        batch_name, batch_frame = self.frame[batch_id]
        case_frame = batch_frame[case_id]

        case_frame_dict = dict()
        for sub_case,sub_case_key in zip(case_frame,['subject','condition','trial','channel','timepoint']):
            case_frame_dict[sub_case_key+'_name'] = sub_case[0]

        return case_frame_dict

    def get_dataframe(self, batch_id=0, case_id=0, average=True, to_print=False):
        batch_name, batch_frame = self.frame[batch_id]
        case_frame = batch_frame[case_id]

        if average:
            data_to_extracted = self.data.average
        else:
            data_to_extracted = self.data.all
        result = group.generate_case_data(case_frame, data_to_extracted)

        if to_print:
            print(f'batch_name: {batch_name}')
            print(f'case_name: {result.name}')
        return result

    def get_array(self, batch_id=0, case_id=0, average=True, to_print=False):
        df = self.dataframe(batch_id, case_id, average, to_print)
        return df.as_matrix()

    def get_index(self, batch_id=0, case_id=0, average=True, to_print=False):
        df = self.dataframe(batch_id, case_id, average, to_print)
        return df.index, df.columns

    def get_info(self, key):
        return self.info[key]

    def iter(self, mode='average'): # iterate the collection and the corresponding batchs, and apply the analyzing function
        def decorator(func):
            def wrapper(*args, **kw):
                data_to_extracted = getattr(self.data, mode) # e.g. self.data.average
     
                analyzed_collection = []
                for batch_name,batch_frame in self.frame:
                    analyzed_batch = []
                    all_case_names = []

                    for case_frame in batch_frame:
                        case_data = group.generate_case_data(case_frame, data_to_extracted)
                        all_case_names.append(case_data.name)
                        result = func(case_data, *args, **kw)

                        if type(result) is tuple: # if return value is not single
                            analyzed_batch.append(result)
                        else:
                            analyzed_batch.append(tuple([result]))

                    all_case_names = np.array(all_case_names)
                    name = ''
                    for ind,i in enumerate(['subjects','conditions','trials','channels','timepoints']):
                        values = list(np.unique(all_case_names[:,ind]))
                        if values != ['All'] and set(values) != set(self.info[i]['all']):
                            name += ','.join(values) + ' '

                    analyzed_batch_dfs = [] 
                    
                    # deal with the multiple return value(s) in the 'func', 
                    # 'len(analyzed_batch[0])' refers to the number of return value(s)
                    for i in range(len(analyzed_batch[0])):
                        analyzed_batch_df = pd.concat([result[i] for result in analyzed_batch])
                        analyzed_batch_df.name = name
                        analyzed_batch_dfs.append(analyzed_batch_df)

                    analyzed_collection.append(analyzed_batch_dfs)

                if len(analyzed_collection[0]) > 1:
                    return tuple([data[i] for data in analyzed_collection] for i in range(len(analyzed_collection[0])))
                else:
                    return [data[0] for data in analyzed_collection]

            return wrapper
        return decorator

class Analyzed_data():
    def __init__(self, analysis_name, data, annotation=None, default_plot_params=dict()):
        for i in data:
            if 'time_group' in i.columns.names and len(i.columns.get_level_values('time_group').unique()) == 1:
                i.columns = i.columns.get_level_values('time')

        self.analysis_name = analysis_name
        self.data = data

        if annotation:
            self.annotation = annotation
        else:
            self.annotation = [None] * len(data)
        self.default_plot_params = default_plot_params

    def correct(self, on_annotation=False, method='fdr_bh'):
        from .statistics.stats_methods import multiple_comparison_correction
        if on_annotation:
            return Analyzed_data(self.analysis_name, self.data, multiple_comparison_correction(self.annotation, method=method), self.default_plot_params)
        else:
            return Analyzed_data(self.analysis_name, multiple_comparison_correction(self.data, method=method), self.annotation, self.default_plot_params)

    def save(self, filepath):
        io.save_result(self, filepath)





from ..default import *
from .. import structure
from .basic import * 
from ..statistics import stats_methods

comparison_params = dict(test=stats_methods.t_test, win='20ms', method='mean', sig_limit=0.05, need_fdr=False)

def ERP(self, compare=False, comparison_params=comparison_params):
    # with the decorator, we can just focuse on case data instead of batch/collection data
    @self.iter('average')
    def to_erp(case_raw_data):
        return case_raw_data.mean(level=['subject','condition_group','channel_group'])
        
    erp_collection = to_erp()
    if compare:
        stats_data = [stats_compare(erp_batch, comparison_params, levels='time', between='condition_group', in_group='subject') 
                    for erp_batch in erp_collection]
    else:
        stats_data = None

    default_plot_params = dict(plot_type=['direct','waveform'], y_title='Amplitude(uV)', err_style='ci_band', color="Set1", style='darkgrid',win=comparison_params['win'],sig_limit=0.05)
    return structure.Analyzed_data('ERP', erp_collection, stats_data, default_plot_params)

def topo_ERPs(self, compare=False, comparison_params=comparison_params):
    # with the decorator, we can just focuse on case data instead of batch/collection data
    @self.iter('average')
    def to_erp(case_raw_data):
        return case_raw_data.mean(level=['subject','condition_group','channel_group'])
    
    erp_collection = to_erp()
    if compare:
        stats_data = [stats_compare(erp_batch, comparison_params, levels=['time','channel_group'], between='condition_group', in_group='subject') 
                    for erp_batch in erp_collection]
    else:
        stats_data = None

    default_plot_params = dict(plot_type=['float','waveform'], y_title='Amplitude(uV)', err_style='ci_band', color="Set1", style='darkgrid',
        win=comparison_params['win'],sig_limit=0.05,xy_locs=self.info['xy_locs'])
    return structure.Analyzed_data('topo_ERPs', erp_collection, stats_data, default_plot_params)

def ERPs(self):
    # with the decorator, we can just focuse on case data instead of batch/collection data
    @self.iter('average')
    def to_erp(case_raw_data):
        return case_raw_data.mean(level=['subject','condition_group','channel_group'])
    
    erp_collection = to_erp()
    stats_data = None

    default_plot_params = dict(plot_type=['direct','waveform'], y_title='Amplitude(uV)', err_style=None, color=['#34495e'], style='darkgrid',legend=False)
    return structure.Analyzed_data('ERPs', erp_collection, stats_data, default_plot_params)

def RMS(self, compare=False, comparison_params=comparison_params):
    def calc_rms(x):
        return np.sqrt(np.mean(x**2))

    # with the decorator, we can just focuse on case data instead of batch/collection data
    @self.iter('average')
    def to_rms(case_raw_data):
        erp = case_raw_data.mean(level=['subject','condition_group','channel','channel_group'])

        RMS_df = []
        for name,data in erp.groupby(level=['subject','condition_group','channel_group']):
            data = pd.DataFrame(data.apply(calc_rms)).T
            data.index = pd.MultiIndex.from_tuples([name], names=['subject','condition_group','channel_group'])
            RMS_df.append(data)
        
        RMS_df = pd.concat(RMS_df)
        return RMS_df
    
    rms_collection = to_rms()
    if compare:
        stats_data = [stats_compare(rms_batch, comparison_params, levels='time', between='condition_group', in_group='subject') 
                        for rms_batch in rms_collection]
    else:
        stats_data = None

    default_plot_params = dict(plot_type=['direct','waveform'], y_title='RMS', err_style='ci_band', color="Set1", style='darkgrid',win=comparison_params['win'],sig_limit=0.05)
    return structure.Analyzed_data('RMS', rms_collection, stats_data, default_plot_params)



from ..default import *

import math
from pylab import get_cmap
from matplotlib.patches import Rectangle

def cmap_discretize(cmap, N):
    if type(cmap) == str:
        cmap = get_cmap(cmap)
    colors_i = np.concatenate((np.linspace(0.1, 1., N), (0.,0.,0.,0.)))
    colors_rgba = cmap(colors_i)
    indices = np.linspace(0, 1., N+1)
    cdict = {}
    for ki,key in enumerate(('red','green','blue')):
        cdict[key] = [ (indices[i], colors_rgba[i-1,ki], colors_rgba[i,ki]) for i in range(N+1) ]
    
    # Return colormap object.
    return matplotlib.colors.LinearSegmentedColormap(cmap.name + "_%d"%N, cdict, 1024)

def refine_axis(ax, title, xticklabels, plot_params):
    xticks = ax.get_xticks()
    #  or len(xticklabels) > 40
    if len(xticks) >= 40:
        if type(xticklabels) is pd.MultiIndex:
            xticklabels = xticklabels.get_level_values('time')

        if xticklabels[-1] - xticklabels[0] < 500:
            step = 50
        else:
            step = 100
        step_num = len(np.arange(xticklabels[0], xticklabels[-1]+5, step))

        xticks = ax.get_xticks()
        xticks = np.linspace(xticks[0],xticks[-1],step_num)
        xticklabels = np.linspace(xticklabels[0],xticklabels[-1],step_num,dtype=int)
        xticklabels_min = xticklabels[0]
        xticklabels_fine = [i//step*step for i in xticklabels]
        xticklabels[0] += 1  # prevent label-xticklabels_min is zero
        xticks_fine = [(label_fine-xticklabels_min)/(label-xticklabels_min)*tick for tick,label,label_fine in zip(xticks,xticklabels,xticklabels_fine)]
        xticks_fine[0] = 0  # prevent xtick_min is less than zero
        
        # padding the last x number in case it lost 
        if xticks[-1] - xticks_fine[-1] > (xticks_fine[-1] - xticks_fine[-2])/2:
            xticks_fine += [xticks_fine[-1] + xticks_fine[-1] - xticks_fine[-2]]
            xticklabels_fine += [xticklabels_fine[-1] + xticklabels_fine[-1] - xticklabels_fine[-2]]
            
        ax.set_xticks(xticks_fine)
        ax.set_xticklabels(xticklabels_fine,rotation=45,fontname='Consolas',size='large')
    else:
        if ax.get_xticklabels()[0].get_text() != ax.get_xticklabels()[1].get_text():
            ax.set_xticklabels(ax.get_xticklabels(),rotation=45,fontname='Consolas',size='large')

    if 'x_title' in plot_params:
        ax.set_xlabel(plot_params['x_title'])
    if 'y_title' in plot_params:
        ax.set_ylabel(plot_params['y_title'])

    if 'title' in plot_params:
        ax.set_title(plot_params['title'])
    else:
        ax.set_title(title)

    if 'ylim' in plot_params:
        ax.set_ylim(plot_params['ylim']) # limitation on y-axis


def heatmap_significant(ax, pv_data, sig_limit=0.05):
    from matplotlib.patches import Rectangle

    significance = [(col_index[1], row_index[0])
                    for row_index, row in pv_data.iterrows() for col_index, cell in row.iteritems()
                    if cell < sig_limit]
    # print(significance)
    for x, y in significance:
        ax.add_patch(Rectangle((pv_data.columns.get_level_values('time').max() - x, 
                                pv_data.index.get_level_values('freq').max() - y), 
                                2, 1, alpha=0.5))




def significant(ax, pv_data, win, sig_limit=0.05): 
    win = int(win[:-2])
    xmin, xmax, ymin, ymax = ax.axis()

    sig_plot_list = np.zeros(int(xmax-xmin))

    for limit,alpha in zip([sig_limit,0.05,0.01,0.001],[0.1,0.2,0.3,0.4]):
        for tp,pv_df in pv_data.items():
            # ugly code
            if type(tp) is tuple:
                tp = tp[1]

            pv = pv_df.values[0]
            if pv and pv<limit<=sig_limit:
                sig_plot_list[int(tp)-win//2:int(tp)+win//2] = alpha

    for tp,alpha in enumerate(sig_plot_list):
        if alpha>0:
            ax.axvspan(tp,tp+1, facecolor='0', alpha=alpha, edgecolor='none')

    cbar_ax = ax.get_figure().add_axes(
        [0.95, 0.4, 0.01, 0.2])  # [left, bottom, width, height]

    cmaplist = [(.2, .2, .2, 1.0), (.3, .3, .3, 1.0), (.4, .4, .4, 1.0)]
    cmap = matplotlib.colors.LinearSegmentedColormap.from_list(
        'Custom cmap', cmaplist, 3)

    # define the bins and normalize
    bounds = np.linspace(0, 3, 4)
    norm = matplotlib.colors.BoundaryNorm(bounds, 3)
    ticks = [(bounds[i]+bounds[i+1])/2 for i in range(len(bounds)-1)]

    cb = matplotlib.colorbar.ColorbarBase(
        cbar_ax, cmap=cmap, norm=norm, spacing='proportional', ticks=ticks, boundaries=bounds, format='%1i')

    cbar_ax.set_title('pvalue')
    cbar_ax.set_yticklabels(['<0.05', '<0.01', '<0.001'])

def plot_waveform(ax, data, plot_params):
    flat_data = pd.DataFrame(data.stack('time'))
    # flat_data.reset_index(level=['condition_group','time'], inplace=True)
    flat_data.reset_index(level=flat_data.index.names, inplace=True)
    
    
    if 'condition_group' in flat_data.columns:
        flat_data['condition_group'] = flat_data['condition_group'].apply(lambda x: ' '.join(x.split(' ')[1:]))
    if 'channel_group' in flat_data.columns:
        flat_data['channel_group'] = flat_data['channel_group'].apply(lambda x: ' '.join(x.split(' ')[1:]))
    
    if 'channel_group' in flat_data.columns and len(flat_data['channel_group'].unique())>1:
        group = "channel_group"
    else:
        group = "condition_group"

    if 'subject' not in flat_data.columns:
        flat_data['subject'] = 0

    if len(flat_data[group].unique()) < 10:
        legend = True
    else:
        legend = False

    if 'err_style' not in plot_params:
        plot_params['err_style'] = None

    sns.tsplot(time="time", value=0, unit='subject', condition=group, data=flat_data, ax=ax, 
        err_style=plot_params['err_style'], legend=legend, color=plot_params['color'])

    xticklabels = flat_data.columns

    refine_axis(ax, data.name, xticklabels, plot_params)

def plot_spectrum(ax, data, plot_params):
    flat_data = data.stack('frequency')
    flat_data.reset_index(level=['condition_group','channel_group','subject','frequency'], inplace=True)
    flat_data['condition_group'] = flat_data['condition_group'].apply(lambda x: ' '.join(x.split(' ')[1:]))
    flat_data['channel_group'] = flat_data['channel_group'].apply(lambda x: ' '.join(x.split(' ')[1:]))
    
    if len(flat_data['channel_group'].unique())>1:
        group = "channel_group"
        legend = False
    else:
        group = "condition_group"
        legend = True

    sns.tsplot(time="frequency", value=0, unit="subject", condition=group, 
        data=flat_data, ax=ax, 
        err_style=plot_params['err_style'], legend=legend, color=plot_params['color'])

    xticklabels = flat_data.columns

    refine_axis(ax, data.name, xticklabels, plot_params)

def plot_heatmap(ax, data, plot_params):
    cbar_ax = ax.get_figure().add_axes([0.95,0.4,0.01,0.2]) # [left, bottom, width, height]
    name = data.name
    if 're_assign' in plot_params:
        # if 'cbar_values' not in plot_params:
        #     plot_params['cbar_values'] = plot_params['re_assign'][0]
        def re_assign(v):
            scale = plot_params['re_assign']
            for idx, new_v in enumerate(scale[1]):
                if (v >= scale[0][idx]) and (v < scale[0][idx+1]):
                    return new_v
        data[data==1] = 1-1e-6
        data = data.applymap(lambda v: re_assign(v))
        
    data.name = name

    if data.index.name and '_group' in data.index.name:
        data.index = [' '.join(i.split(' ')[1:]) for i in data.index]

    if 'cbar_values' in plot_params:
        cmap=cmap_discretize(plot_params['color'],len(plot_params['cbar_values']))
        sns.heatmap(data,ax=ax,cbar_ax=cbar_ax,cmap=cmap)
        cb_yticks = cbar_ax.get_yticks()
        cbar_ax.yaxis.set_ticks([(cb_yticks[i]+cb_yticks[i+1])/2*1.5 for i in range(len(cb_yticks)-1)])
        cbar_ax.set_yticklabels(plot_params['cbar_values'])
    else:
        sns.heatmap(data,ax=ax,cbar_ax=cbar_ax,cmap=plot_params['color'])

    xticklabels = data.columns

    refine_axis(ax, data.name, xticklabels, plot_params)

    if len(xticklabels)<40:
        ax.set_aspect(1) # ratio between y_unit and x_unit
    else:
        ax.set_aspect(len(xticklabels)/40)

    ax.set_yticklabels(ax.get_yticklabels(),rotation=0,fontname='Consolas')

    cbar_ax.set_title(plot_params['cbar_title'])

    if 'grid' in plot_params and plot_params['grid']:
        for i in ax.get_xticks():
            ax.axvline(i, c="w",linewidth=1,linestyle='dotted')
        for i in range(len(data.index)):
            ax.axhline(i, c="w",linewidth=2)

def get_topograph(data, locations, channels, N):
    locs = np.array([locations[chan] for chan in channels])

    x_span = locs[:,1].max()-locs[:,1].min()
    y_span = locs[:,0].max()-locs[:,0].min()
    xi = np.linspace(0, x_span, N)[None,:]
    yi = np.linspace(0, y_span, N)[:,None]

    locs = [(x-locs[:,1].min(),y-locs[:,0].min()) for y,x in locs]
    locs.extend([(0,0), (0,y_span), (x_span,0), (x_span,y_span)])

    return scipy.interpolate.griddata(locs, np.append(data, [0,0,0,0]), (xi, yi), method='cubic')

def plot_topograph(ax, data, plot_params):
    amp_data = pd.DataFrame(data['Amp'])

    channel_index = data['Amp'].index.get_level_values('channel')

    amp_data['y_locs'] = [plot_params['chan_locs'][ch][0] for ch in channel_index]
    amp_data['x_locs'] = [plot_params['chan_locs'][ch][1] for ch in channel_index]

    # http://stackoverflow.com/questions/15361143/how-to-fit-result-of-matplotlib-pyplot-contourf-into-circle

    N = 50    # size for interpolation, small number will make a dentate edge

    bottom,top = amp_data['y_locs'].min()-0.1, amp_data['y_locs'].max()+0.1 # set margin in the plot
    left,right = bottom,top
    corners = pd.DataFrame([[0,left,bottom], [0,right,bottom], [0,left,top], [0,right,top]],columns=['Amp','x_locs','y_locs'])
    amp_data = amp_data.append(corners)

    x,y,z = amp_data['y_locs']-left, amp_data['x_locs']-bottom, amp_data['Amp']

    xi = np.linspace(0,right-left, N)
    yi = np.linspace(0,top-bottom, N)
    zi = scipy.interpolate.griddata((x,y), z, (xi[None,:], yi[:,None]), method='cubic')

    # zi = get_topograph(data, plot_params['chan_locs'], channel_index, N)

    if 'zlim' in plot_params:
        zlim = plot_params['zlim']
        zi[zi<zlim[0]] = zlim[0]
        zi[zi>zlim[1]] = zlim[1]

    if 'mask' not in plot_params or plot_params['mask']:
        # make a circle mask
        # set points > radius to not-a-number. They will not be plotted.
        # the dr/2 makes the edges a bit smoother
        xy_center = [(right - left)/2,(top - bottom)/2]   # center of the plot
        radius = (amp_data['x_locs'].max() - amp_data['x_locs'].min())/2 + 0.05
        dr = xi[1] - xi[0]
        for i in range(N):
            for j in range(N):
                r = np.sqrt((xi[i] - xy_center[0])**2 + (yi[j] - xy_center[1])**2)
                if (r - dr/2) > radius:
                    zi[j,i] = "nan"

        # draw a circle
        # change the linewidth to hide the 
        circle = matplotlib.patches.Circle(xy = xy_center, radius = radius-0.5, edgecolor = "k", facecolor = "none")
        ax.add_patch(circle)
        # draw a nose
        points = [(xy_center[0]-(radius-0.5)/5, xy_center[1]+radius-0.5), (xy_center[0], xy_center[1]+radius), (xy_center[0]+(radius-0.5)/5, xy_center[1]+radius-0.5)]
        line = plt.Polygon(points, closed=None, fill=None, edgecolor = "k", facecolor = "none")
        ax.add_patch(line)

    # # draw ears
    # points = [(xy_center[0]-(radius-0.5), xy_center[1]+(radius-0.5)/5), (xy_center[0]-radius, xy_center[1]), (xy_center[0]-(radius-0.5), xy_center[1]-(radius-0.5)/5)]
    # line = plt.Polygon(points, closed=None, fill=None, edgecolor = "k", facecolor = "none")
    # ax.add_patch(line)
    # points = [(xy_center[0]+(radius-0.5), xy_center[1]+(radius-0.5)/5), (xy_center[0]+radius, xy_center[1]), (xy_center[0]+(radius-0.5), xy_center[1]-(radius-0.5)/5)]
    # line = plt.Polygon(points, closed=None, fill=None, edgecolor = "k", facecolor = "none")
    # ax.add_patch(line)

    # use different number of levels for the fill and the lines
    if 'zlim' in plot_params:
        zlim = plot_params['zlim']
        CS = ax.contourf(xi, yi, zi, 60, cmap = plot_params['color'], zorder = 1,levels=np.linspace(zlim[0],zlim[1],40))
    else:
        CS = ax.contourf(xi, yi, zi, 60, cmap = plot_params['color'], zorder = 1)
    # contour line
    ax.contour(xi, yi, zi, colors = "grey", alpha = 0.5, zorder = 2, linestyles='solid')

    # add the data points, x[:-4] will drop the corner points we just added 
    ax.scatter(x[:-4], y[:-4], marker = 'o', c = 'b', alpha = 0.5, s = 2, zorder = 3)  

    # draw significant channel as white point
    if 'p_val' in data.columns:
        pval_data = pd.DataFrame(data['p_val'])
        channel_index = data['p_val'].index.get_level_values('channel')
        pval_data['y_locs'] = [plot_params['chan_locs'][ch][0] for ch in channel_index]
        pval_data['x_locs'] = [plot_params['chan_locs'][ch][1] for ch in channel_index]

        for ind,i in pval_data.iterrows():
            if i.p_val<plot_params['sig_limit']:
                ax.scatter(i.y_locs-bottom, i.x_locs-left,  c = 'w', s = 25, alpha=1, zorder = 3, linewidths=0.5, edgecolors='b')

    # remove the axis
    ax.axis('off')

def channel_locs(topo):
    # epochs.info['xy_locs']

    label = [i for i in topo.keys()]
    x = [i[0] for i in topo.values()]
    y = [i[1] for i in topo.values()]

    fig, ax = plt.subplots()
    ax.scatter(x, y)
    for i, txt in enumerate(label):
        ax.annotate(txt, (x[i],y[i]))

    ax.axis('off')
    plt.show()



from . import basic

from .erp import *
from .spectrum import *
from .topo import *
from .cluster_models import *
from .classification_models import *
from .cosine_distance_models import *

# from .other import *

structure.Extracted_epochs.ERP = ERP
structure.Extracted_epochs.topo_ERPs = topo_ERPs
structure.Extracted_epochs.ERPs = ERPs
structure.Extracted_epochs.RMS = RMS
structure.Extracted_epochs.Spectrum = Spectrum
structure.Extracted_epochs.Time_frequency = Time_frequency
structure.Extracted_epochs.topography = topography  
structure.Extracted_epochs.frequency_topography = frequency_topography
structure.Extracted_epochs.significant_channels_count = significant_channels_count
structure.Extracted_epochs.clustering = clustering
structure.Extracted_epochs.tanova = tanova
structure.Extracted_epochs.cosine_distance_dynamics = cosine_distance_dynamics
structure.Extracted_epochs.classification = classification



from .default import *

Case = namedtuple('Case',['name','val'])
Batch = namedtuple('Batch',['name','val'])

def parsing(collection_script, epochs_data):
    def conditions_filter_by_factor(query_str):
        cond_format = epochs_data.info['conditions']['format']
        # if there is anything between brackets
        if len(re.sub(r'\(.*?\)','',cond_format))>0: 
            cond_format = cond_format.replace('(','(?P<c_').replace(')','>.*)')
        else:
            cond_format = cond_format.replace('(','(?P<c_').replace(')','>.)')
        # generate the dataframe of factors    
        condition_df = []
        for condition in epochs_data.info['conditions']['all']:
            pattern = re.search(cond_format, condition)
            if pattern:
                indicators = pattern.groupdict()
                indicators['name'] = condition
                condition_df.append(indicators)
        if condition_df!=[]:
            condition_df = pd.DataFrame(condition_df)
        else:
            raise Exception(f'Cannot find any thing by the format "{epochs_data.info["conditions"]["format"]}"')
        query_str = query_str.replace('|','||').replace('&','&&')+'$'
        query_str = re.sub(r'(^|\W)(\w*?)!=(.*?)([\W])',r'\1c_\2!="\3"\4',query_str)
        query_str = re.sub(r'(^|\W)([ a-zA-Z0-9]*?)=(.*?)([\W])',r'\1c_\2=="\3"\4',query_str)
        query_str = re.sub(r'!c_==""',r'!=',query_str)
        query_str = query_str[:-1].replace('||','|').replace('&&','&')
        return condition_df.query(query_str).name.tolist()

    def group_by_list(li,levels):
        # li = ['aB1','aC1','bB1','bC1','aB2','aC2','bB2','bC2']
        # group_by_list(li,[1,2])
        
        def groupBy(li,level):
            li.sort(key=lambda x:x[level])
            return [list(sub_li) for i, sub_li in itertools.groupby(li,lambda x:x[level])]
        
        if len(levels)>0:
            return [group_by_list(sub_li,levels[1:]) for sub_li in groupBy(li,levels[0])]
        else:
            return li

    def case_parser(key, case_str):
        def ambigous_symbol_converter(old_str):
            while 1:
                new_str = re.sub(r'{([^}]*?)\&(.*?)}', '{\g<1><and>\g<2>}',old_str)
                new_str = re.sub(r'{([^}]*?)\|(.*?)}', '{\g<1><or>\g<2>}',new_str)
                if old_str==new_str:
                    return new_str
                else:
                    old_str = new_str

        def plus_minus(text): # make the PLUS mix and MINUS mix
            text = '+' + text.replace(' ','')
            mix = re.findall(r'[\+\-]{,2}[^\+\-]+',text)
            plus,minus=[],[]
            for c in mix:   
                if c[0]=='+':
                    plus.append(c[1:])
                elif c[0]=='-':
                    minus.append(c[1:])
            if plus==[]: plus = slice(None)
            return {'+':plus,'-':minus} 

        def timepoint_parser(timepoint_str_list):
            interval = 1000//epochs_data.info['sample_rate']
            timepoint_list = []
            for timepoint_str in timepoint_str_list:
                if '~' in timepoint_str:
                    start,end = timepoint_str.split('~')
                    timepoint_list += list(np.arange(int(start),int(end)+interval,interval).astype(int))
                else:
                    timepoint_list.append(int(timepoint_str))

            return timepoint_list

        def subject_parser(subject_str_list):
            subject_list = [int(i) for i in subject_str_list]
            return subject_list

        def trial_parser(trial_str_list):
            trial_list = []
            for trial_str in trial_str_list:
                if '~' in trial_str:
                    start,end = trial_str.split('~')
                    trial_list += list(range(int(start),int(end)+1))
                else:
                    trial_list.append(int(trial_str))
            return trial_list

        def condition_parser(condition_str_list):
            # epochs.info['conditions']['format'] = '(global)_(local)_(target)'
            # 'a:{global=T&target=G}&{global=F&target=G}'

            condition_list = []
            for condition_str in condition_str_list:
                if condition_str[0] is '{' and condition_str[-1] is '}' :
                    query_str = condition_str[1:-1].replace('<and>','&').replace('<or>','|')
                    condition_str = conditions_filter_by_factor(query_str)
                else:
                    condition_str = [condition_str]
                condition_list += condition_str
            return condition_list

        if ':' in case_str:
            name,val = case_str.split(':')
        else:
            name,val = case_str,case_str

        val = ambigous_symbol_converter(val)
        val = [plus_minus(i) for i in val.split('&')]
        
        if key=='timepoints':
            val = [{'+':timepoint_parser(part['+']),'-':timepoint_parser(part['-'])}
                    for part in val]
        elif key=='trials':
            val = [{'+':trial_parser(part['+']),'-':trial_parser(part['-'])}
                        for part in val]
        elif key=='conditions':
            val = [{'+':condition_parser(part['+']),'-':condition_parser(part['-'])}
                        for part in val]
        elif key=='subjects':
            val = [{'+':subject_parser(part['+']),'-':subject_parser(part['-'])}
                        for part in val]
        return Case(name,val)

    def batch_str_parser(batch_str):
        batch = dict()
        batch_str_list = batch_str.split('@')

        batch['conditions'] = batch_str_list[0].strip()
        if len(batch_str_list)>0:
            chs = batch_str_list[1].strip()
            if chs!='':
                batch['channels'] = chs
        if len(batch_str_list)>1:
            tps = batch_str_list[2].strip()
            if tps!='':
               batch['timepoints'] = tps
        return batch

    def combine(key,sep):
        li = epochs_data.info[key]['all']
        if sep=='+':
            return ['All:' + sep.join([str(i) for i in li])]
        elif sep==',':
            return [str(i) for i in li]

    if type(collection_script) in [str,dict]:
        collection_script = [collection_script]
    collection_frame = []
    # iter the collection
    for batch_script in collection_script:
        # convert the description into dict if the description is a string
        if type(batch_script) is str:
            batch_script = batch_str_parser(batch_script)
        elif type(batch_script) is not dict:
            raise Exception(f'The description "{batch_script}" should be a string or dict!')
        
        # convert into full-description dict
        batch_script_full = dict()
        for key in ['subjects', 'conditions', 'trials', 'channels', 'timepoints']:
            if key in batch_script:
                if batch_script[key] == 'each':
                    batch_script_full[key] = combine(key,',')
                else:
                    if type(batch_script[key]) is str:
                        batch_script_full[key] = batch_script[key].split(',')
                    else:
                        batch_script_full[key] = batch_script[key]
            else:
                batch_script_full[key] = combine(key,'+')

            # parse the string of case
            batch_script_full[key] = [case_parser(key,i) for i in batch_script_full[key]]
        
        batch_name = []
        for batch_k,batch_v in batch_script_full.items():
            batch_desp = ','.join([case_v.name for case_v in batch_v])
            if batch_desp!='All':
                batch_name.append(batch_desp)

        # Get the cartesian product of the dict
        batch_frame = []
        for case_frame in list(itertools.product(*batch_script_full.values())):
            batch_frame.append(case_frame)

        collection_frame.append(Batch(batch_name, batch_frame))

    return collection_frame

def filter(epochs_data, subject=slice(None), condition=slice(None), trial=slice(None), channel=slice(None), timepoint=slice(None), filter_dict=dict()):
    if 'subject' in filter_dict:
        subject = filter_dict['subject']
    if 'condition' in filter_dict:
        condition = filter_dict['condition']
    if 'trial' in filter_dict:
        trial = filter_dict['trial']
    if 'channel' in filter_dict:
        channel = filter_dict['channel']
    if 'timepoint' in filter_dict:
        timepoint = filter_dict['timepoint']

    return epochs_data.loc[ids[subject,condition,trial,channel],ids[timepoint]]

def generate_case_data(case_frame, epochs_data_todo):
    case_frame_dict = dict()
    for sub_case,sub_case_key in zip(case_frame,['subject','condition','trial','channel','timepoint']):
        case_frame_dict[sub_case_key+'_name'] = sub_case[0]
        case_frame_dict[sub_case_key] = sub_case[1]

    get_all = lambda l: sum([i['+']+i['-'] for i in l], [])
    extracted_data = epochs_data_todo.loc[
            ids[get_all(case_frame_dict['subject']),
                get_all(case_frame_dict['condition']),
                get_all(case_frame_dict['trial']),
                get_all(case_frame_dict['channel'])
               ],ids[get_all(case_frame_dict['timepoint'])]]

    # subtract baselines
    def subtract(data,baseline,level):
        if target_level == 'timepoint':
            baseline = baseline.mean(axis=1)
            return data.subtract(baseline,axis=0)
        else:
            name_d = {'subject':0,'condition':0,'trial':0,'channel':0}

            levels_keeped = ['subject','condition','channel']
            if 'trial' != level:
                levels_keeped.remove(level)

            baseline = baseline.mean(level=levels_keeped)
            baseline_dict = dict((ind,list(i)) for ind,i in baseline.iterrows())
            
            def sub(x):
                name_d['subject'],name_d['condition'],name_d['trial'],name_d['channel'] = x.name
                ind = tuple(name_d[i] for i in levels_keeped)
                return x-baseline_dict[ind]

            return data.apply(sub,axis=1)

    levels = ['subject','condition','trial','channel','timepoint']
    for target_level in levels:
        data_t = []
        for group_idx,group in enumerate(case_frame_dict[target_level]):
            group_data = filter(extracted_data, filter_dict={target_level:group['+']})

            if len(group['-'])>0:
                baseline = filter(extracted_data, filter_dict={target_level:group['-']})
                group_data = subtract(group_data, baseline, target_level)

            if target_level == 'timepoint':
                group_data.columns = pd.MultiIndex.from_tuples([(group_idx,i) for i in group_data.columns])
            else:
                group_data.index.set_levels(
                                [f'{group_idx} {i}' for i in group_data.index.levels[levels.index(target_level)]],
                                        target_level,inplace=True)

            data_t.append(group_data)

        extracted_data = pd.concat(data_t)
        extracted_data.sort_index(inplace=True)
    
    for target_level in levels[:-1]:
        # extracted_data[target_level+'_group'] = [i.split(' ')[0] for i in extracted_data.index.get_level_values(target_level)]
        extracted_data[target_level+'_group'] = [i.split(' ')[0]+' '+case_frame_dict[target_level+'_name'] for i in extracted_data.index.get_level_values(target_level)]
        extracted_data.set_index(target_level+'_group', append=True, inplace=True)
        extracted_data.columns.set_levels(list(extracted_data.columns.levels[1][:-1]), level=1, inplace=True)
        extracted_data.index.set_levels(
                    [' '.join(i.split(' ')[1:]) for i in extracted_data.index.levels[levels.index(target_level)]], 
                    target_level,inplace=True)

    extracted_data.columns.names = ['time_group','time']
    extracted_data.name = [case_frame_dict[target_level+'_name'] for target_level in levels]
    return extracted_data

# '''to be deprecated'''
# def extract(epochs, collection_script, average=False, preload=False):
#     collection_frame = parsing(collection_script, epochs)    

#     if average:
#         epochs_data_todo = epochs.average
#     else:
#         epochs_data_todo = epochs.all

#     collection_data = []
#     for batch_name,batch_frame in collection_frame:
#         batch_data = (generate_case_data(case_frame, epochs_data_todo)
#                         for case_frame in batch_frame)
#         if preload:
#             batch_data = list(batch_data)
#         collection_data.append(Batch(batch_name, batch_data))

#     return collection_data


from ..default import *
from tqdm import tqdm
from ..statistics import stats_methods
import numexpr as ne

# todo: group single or not, 2*2 or not
# def check_availability(data,single_value_level=[],complex_value_level=[]):
#     for l in single_value_level:
#         if l == 'time_group':
#             values = list(data.columns.get_level_values(l).unique())
#         else:
#             values = list(data.index.get_level_values(l).unique())
#         if len(values)>1:
#             raise Exception(f'level "{l}" should have only one value, but {values} in there.')
#     for l in complex_value_level:
#         if l == 'time_group':
#             values = list(data.columns.get_level_values(l).unique())
#         else:
#             values = list(data.index.get_level_values(l).unique())
#         if len(values)==1:
#             raise Exception(f'level "{l}" should have more than one value, but {values} in there.')

def check_availability(data, level, condition):
    if level == 'time_group':
        values = list(data.columns.get_level_values(level).unique())
    else:
        values = list(data.index.get_level_values(level).unique())
    target = len(values)
    if not ne.evaluate(str(target)+condition):
        raise Exception(
            f'level "{level}"\'s unique_value_count should {condition} , but {values} in there.')

def shuffle_on_level(df, level, within_subject=True, inplace=True):
    raw = list(zip(df.index.get_level_values('subject'),df.index.labels[df.index.names.index(level)],df.index.get_level_values('trial')))
    if within_subject:
        unique = list(set(raw)) # unique ind
        unique.sort(key=lambda x:x[0]) #需要先排序，然后才能groupby
        g = itertools.groupby(unique,lambda x:x[0])
        mapping = dict()
        for key,ind in g:
            ind = list(ind)
            new_ind = list(zip([i[0] for i in ind],np.random.permutation([i[1] for i in ind]),[i[2] for i in ind]))
            mapping.update(dict(zip(ind,new_ind)))
        group_labels = [mapping[i][1] for i in raw]
        # within_labels = df.index.get_level_values(within)
        # group_labels = df.index.get_level_values(level)
        # label_groups = dict()
        # for idx,(within_key,group_key) in enumerate(zip(within_labels, group_labels)):
        #     if within_key not in label_groups:
        #         label_groups[within_key] = [[],[]]
        #     label_groups[within_key][0].append(ind)
        #     label_groups[within_key][0].append(group_key)

        # group_labels = []
        # for within_key,(inds,group_keys) in label_groups.items():
        #     random.shuffle(group_keys)
        #     group_labels.extend(list(zip(inds,group_keys)))
        # group_labels.sort(key=lambda x:x[0])
        # group_labels = [i[1] for i in group_labels]
    else:
        ind = list(set(raw)) # unique ind
        new_ind = list(zip([i[0] for i in ind],np.random.permutation([i[1] for i in ind]),[i[2] for i in ind]))
        mapping = dict(zip(ind,new_ind))
        group_labels = [mapping[i][1] for i in raw]

    if inplace:
        # print(level,'\n',group_labels)
        df.index = df.index.set_labels(group_labels,level=level)
    else:
        df_new = df.copy()
        df_new.index = df.index.set_labels(group_labels,level=level)
        return df

def add_index(df, name, value):
    df[name] = value
    df.set_index(name, append=True, inplace=True)

def recover_index(df, old_df, name):
    if '_group' in name:
        values = old_df.index.get_level_values(name).map(lambda x: '0 '+' '.join(x.split(' ')[1:])).unique()
    else:
        values = old_df.index.get_level_values(name).unique()
    if len(values)==1: # make sure if the value in old_df index is unique'
        value = values[0]
    else:
        value = ','.join(values)
        # raise Exception(f'The values in level {name} should be unique')
    add_index(df, name, value)

def average(df, keep=None, remove=None):
    if keep==None:
        if remove==None:
            raise Exception('You should provide the value to "keep" or "remove"')
        else:
            all_levels = df.index.names
            keep = list(set(all_levels)-set(remove))
    return df.mean(level=keep)

'sampling'
def point_sampling(df, step_size):
    if step_size[-2:] in ['ms','Ms','MS'] and step_size[:-2].isdigit():
        step_size = step_size[:-2]+'N'
    else:
        raise Exception('step_size should be a string formatted as "10ms" or "2ms"')

    new_df = []
    for idx,group_data in df.groupby(axis=1,level='time_group'):
        old_tps = [pd.Timedelta(tp) for tp in group_data.columns.levels[1]]
        tp_indexs = [old_tps.index(tp) for tp in pd.timedelta_range(old_tps[0], old_tps[-1], freq=step_size) if tp in old_tps]
        group_data = group_data.iloc[:,tp_indexs]
        new_df.append(group_data)
    new_df = pd.concat(new_df,axis=1)

    return new_df

def window_sampling(df, win_size, sample='mean'):
    if win_size[-2:] in ['ms','Ms','MS'] and win_size[:-2].isdigit():
        win_size = win_size[:-2]+'N'
    else:
        raise Exception('win_size should be a string formatted as "20ms" or "5ms"')

    new_df = []
    for idx,group_data in df.groupby(axis=1,level='time_group'):
        group_data.columns.set_levels([pd.Timedelta(tp) for tp in group_data.columns.levels[1]], level=1, inplace=True)
        group_data_new = group_data.resample(win_size, axis=1, how='mean', level=1)
        group_data_new.columns = group_data_new.columns + pd.Timedelta(win_size[:-1]+'ns')/2 # loffset of `resample` doesn't work, so I add this line 
        if group_data_new.columns[-1]>group_data[0].columns[-1]:
            del group_data_new[group_data_new.columns[-1]]
        group_data_new.columns = pd.MultiIndex.from_product([[idx], group_data_new.columns.values.tolist()],names=['time_group','time'])
        new_df.append(group_data_new)

    new_df = pd.concat(new_df,axis=1)

    return new_df

def sampling(data,step_size='1ms',win_size='1ms',sample='mean'):
    if step_size!='1ms':
        data = point_sampling(data, step_size)
    elif win_size!='1ms':
        data = window_sampling(data, win_size, sample=sample)
    return data

'map and apply'
# divide into units, convert the units, then combine them
def convert(df, unit, func):
    df_t = df.mean(level=unit)
    converted_list = [func(data) for name, data in df_t.groupby(level=unit)]
    return pd.concat(converted_list)

# def roll_on_levels(df, func, arguments_dict=dict(), levels='time', prograssbar=True):
#     col_level = df.columns.names[1]
#     df = df.stack(col_level)
#     result = []
#     group_ids = []

#     for group_id,group_data in tqdm(df.groupby(level=levels), ncols=0, disable=(not prograssbar)):
#         # print(group_data)
#         result_one_group = func(group_data, **arguments_dict)

#         group_ids.append(group_id)
#         result.append(result_one_group)

#     if type(levels)==str or len(levels)==1:
#         index = pd.Index(group_ids, name=levels)
#     else:
#         index = pd.MultiIndex.from_tuples(group_ids, names=levels)

#     result = pd.DataFrame(result, index=index)
#     # result.name = df.name
#     recover_index(result, df, 'condition_group')
#     return result.unstack(col_level)


class OverlappedSequence():
    def __init__(self, df, sequence_size):
        self.df = df
        self.sequence_size = sequence_size
        self.count = len(df.columns)

    def __len__(self):
        return self.count

    def __iter__(self):
        return (self.df.iloc[:, i-self.sequence_size:i] for i in range(self.sequence_size, self.count))


from concurrent.futures import ProcessPoolExecutor
def roll_on_levels(df, func, arguments_dict=dict(), levels='time', prograssbar=True, parallel=False, sequence_size=1):
    col_level = df.columns.names[1]
    df = df.stack(col_level)
    data_results = []
    annotation_results = []
    group_ids = []
    if sequence_size==1:
        groups_to_roll = df.groupby(level=levels)
    else:
        '[beta] for time-sequence models'
        groups_to_roll = OverlappedSequence(df, sequence_size)

    if parallel is False:
        for group_id,group_data in tqdm(groups_to_roll, ncols=0, disable=(not prograssbar)):
            result_one_group = func(group_data, **arguments_dict)

            group_ids.append(group_id)
            if type(result_one_group) is tuple:
                if len(result_one_group)==2:
                    data_results.append(result_one_group[0])
                    annotation_results.append(result_one_group[1])
                else:
                    raise Exception('The return value of func should not above 2!')
            else:
                data_results.append(result_one_group)
    else:
        if parallel is True: parallel = None # if max_workers is None or not given, it will default to the number of processors on the machine
        with ProcessPoolExecutor(max_workers=parallel) as executor:
            tasks = [(group_id, 
                      executor.submit(func, group_data, **arguments_dict))
                    for group_id,group_data in groups_to_roll]
            with tqdm(total=len(tasks), disable=(not prograssbar)) as pbar:  
                while 1:
                    count_doned = sum([task.done() for group_id,task in tasks])
                    pbar.update(count_doned - pbar.n)
                    if count_doned==len(tasks):
                        break

        for group_id,task in tasks:
            group_ids.append(group_id)
            result_one_group = task.result()
            if type(result_one_group) is tuple:
                if len(result_one_group)==2:
                    data_results.append(result_one_group[0])
                    annotation_results.append(result_one_group[1])
                else:
                    raise Exception('The return value of func should not above 2!')
            else:
                data_results.append(result_one_group)

    if type(levels)==str or len(levels)==1:
        index = pd.Index(group_ids, name=levels)
    else:
        index = pd.MultiIndex.from_tuples(group_ids, names=levels)
    
    data_results = pd.DataFrame(data_results, index=index)
    recover_index(data_results, df, 'condition_group')

    if len(annotation_results)==0:
        return data_results.unstack(col_level)
    else:
        annotation_results = pd.DataFrame(annotation_results, index=index)
        recover_index(annotation_results, df, 'condition_group')
        return data_results.unstack(col_level), annotation_results.unstack(col_level)

# ugly code. To do.
def roll_on_levels_and_compare(df, func, arguments_dict=dict(), levels='time', between='condition_group', in_group='subject',prograssbar=True):
    col_level = df.columns.names[1]
    df = df.stack(col_level)
    result = []
    group_ids = []

    for group_id,group_data in tqdm(df.groupby(level=levels), ncols=0, disable=(not prograssbar)):
        data_sub = [group_data.mean(level=in_group).mean(axis=1) 
            for group_id,group_data in group_data.groupby(level=between)]
        
        # ipdb.set_trace()
        result_one_group = func(data_sub, **arguments_dict)['pvalue']
        # re_add_name = group_data.index.get_level_values(between)[0][2:]
        
        group_ids.append(group_id)
        result.append(result_one_group)

    if type(levels)==str or len(levels)==1:
        index = pd.Index(group_ids, name=levels)
    else:
        index = pd.MultiIndex.from_tuples(group_ids, names=levels)

    result = pd.DataFrame(result, index=index)
    
    recover_index(result, group_data, between)

    return result.unstack(col_level)

def stats_compare(df, comparison, levels='time', between='condition_group', in_group='subject'):
    condition_groups = df.index.get_level_values(between).unique()
    if comparison['sig_limit']>0 and len(condition_groups)==2:
        if comparison['win']!='1ms':
            df = window_sampling(df, comparison['win'], comparison['method'])
        stats_result = roll_on_levels_and_compare(df, comparison['test'], levels=levels, between=between, in_group=in_group)
        pval_result = stats_result
        return pval_result
    else:
        return None




import re
import os

path = '../'

'quickstart_generator'
with open(path+'docs/quickstart_template.txt') as f:
    template = f.readlines()

    new_file = []
    for l in template:
        if l[:2] != '& ':
            new_file.append(l)
        else:
            try:
                with open('../test/'+l[2:].replace('\n',''), encoding='utf8') as f:
                    lines = f.readlines()

                    for ind,l in enumerate(lines):
                        if l[:18] == "from expy import *":
                            break

                    code = ''.join(lines[ind:])

                new_file.append('\n```python\n%s\n```\n' %(code))
            except:
                    print('cannnot read',file)


    with open(path+'docs/quickstart.md','w+', encoding='utf8') as f:
        f.writelines(new_file)

'cookbook_generator'
with open(path+'docs/cookbook_template.txt') as f:
    template = f.readlines()

    new_file = []
    for l in template:
        if l[:2] != '& ':
            new_file.append(l)
        else:
            try:
                with open('../test/'+l[2:].replace('\n',''), encoding='utf8') as f:
                    lines = f.readlines()

                    for ind,l in enumerate(lines):
                        if l[:18] == "from expy import *":
                            ind += 1
                            break

                    code = ''.join(lines[ind:])

                new_file.append('\n```python\n%s\n```\n' %(code))
            except:
                    print('cannnot read',file)


    with open(path+'docs/cookbook.md','w+', encoding='utf8') as f:
        f.writelines(new_file)

'api_generator'
with open(path+'docs/api_template.txt') as f:
    template = f.readlines()

    comments = dict()
    for root, _dirs, files in os.walk(path):
        for file in files:
            if file[-3:]=='.py' and root[-4:] != 'test':
                try:
                    with open(root+'\\'+file,encoding='utf8') as f:
                        
                        lines = f.readlines()

                        defines = []
                        for ind,l in enumerate(lines):
                            if  l[:4] == "def ":
                                defines.append([l,ind])

                        for define,ind in defines:
                            if lines[ind+1] == "    '''\n":
                                f = re.search('def (.*?)\((.*)\)',define)
                                f_name = f.group(1)
                                f_args = f.group(2)
                                comments[f_name] = [f_args,'']
                                ind += 2
                                while lines[ind] != "    '''\n":
                                    comments[f_name][1] += (lines[ind][4:-1] + '\n')
                                    ind += 1
                except:
                    print('cannnot read',file)

    new_file = []
    for l in template:
        if '- **' != l[:4]:
            new_file.append(l)
        else:
            f_args,comm = comments[l[4:-3]]
            new_file.append('\n- **%s(%s)**\n\n' %(l[4:-3],f_args))
            new_file.append('\n```\n%s```\n\n' %(comm))

    with open(path+'docs/api.md','w+', encoding='utf8') as f:
        f.writelines(new_file)


from ..default import *
from .. import structure
from .basic import * 
from ..statistics import stats_methods

from scipy.spatial.distance import cosine # 1-cosD

def calc_cosD(df): # df: channel,subject,condition_group
    cond_A, cond_B = [average(conditon_group_data, keep='channel').mean(axis=1) 
        for conditon_group_id,conditon_group_data in df.groupby(level='condition_group')]
    return cosine(cond_A, cond_B)

def sub_func(group_data, shuffle=500, within_subject=True):
    result_real = calc_cosD(group_data)
    dist_baseline = []
    for _ in range(shuffle):
        shuffle_on_level(group_data, 'condition_group', within_subject=within_subject)
        dist_baseline.append(calc_cosD(group_data))

    pvalue = stats_methods.get_pvalue_from_distribution(result_real, dist_baseline)
    return pvalue, result_real

def tanova(self,step_size='1ms',win_size='1ms',sample='mean',shuffle=500,mode=1,parallel=False):
    # with the decorator, we can just focuse on case data instead of batch/collection data
    @self.iter('all')
    def to_tanova1(case_raw_data):
        case_raw_data = sampling(case_raw_data, step_size, win_size, sample)
        check_availability(case_raw_data, 'condition_group', '==2')
        return roll_on_levels(case_raw_data, sub_func, arguments_dict=dict(shuffle=shuffle, within_subject=False), levels='time', prograssbar=True, parallel=parallel)

    @self.iter('all')
    def to_tanova2(case_raw_data):
        case_raw_data = sampling(case_raw_data, step_size, win_size, sample)
        check_availability(case_raw_data, 'condition_group', '==2')
        return roll_on_levels(case_raw_data, sub_func, arguments_dict=dict(shuffle=shuffle, within_subject=True), levels='time', prograssbar=True, parallel=parallel)

    @self.iter('average')
    def to_tanova3(case_raw_data):
        case_raw_data = sampling(case_raw_data, step_size, win_size, sample)
        check_availability(case_raw_data, 'condition_group', '==2')
        return roll_on_levels(case_raw_data, sub_func, arguments_dict=dict(shuffle=shuffle, within_subject=True), levels='time', prograssbar=True, parallel=parallel)

    if mode==1:
        tanova_collection, annotation_collection = to_tanova1()
    elif mode==2:
        tanova_collection, annotation_collection = to_tanova2()
    elif mode==3:
        tanova_collection, annotation_collection = to_tanova3()

    default_plot_params = dict(title='TANOVA',plot_type=['direct','heatmap'], x_len=12, re_assign=[(0,0.01,0.05,0.1,1),(4,3,2,1)],
                                color=sns.cubehelix_palette(light=1, as_cmap=True), grid=True,
                                x_title='time', y_title='condition_group',cbar_title='pvalue',cbar_values=['>0.1','<0.1','<0.05','<0.01'])

    return structure.Analyzed_data('TANOVA', tanova_collection, annotation_collection, default_plot_params=default_plot_params)

def cosine_distance_dynamics(self):
    # with the decorator, we can just focuse on case data instead of batch/collection data
    @self.iter('average')
    def calc(case_raw_data):
        check_availability(case_raw_data, 'condition_group', '==2')
        return roll_on_levels(case_raw_data, calc_cosD, levels='time')

    cosine_distance_collection = calc()

    default_plot_params = dict(title='cosine_distance_dynamics', plot_type=['direct','waveform'], x_title='time', y_title='distance', color="Set1", style='darkgrid')
    return structure.Analyzed_data('cosine distance dynamics', cosine_distance_collection, default_plot_params=default_plot_params)



from ..default import *
from .. import structure
from . import figure_unit

import matplotlib.gridspec as gridspec

def plot(self, plot_params=None, save=False, return_fig=False):
    'setting'
    if plot_params:
        for k in self.default_plot_params:
            if k not in plot_params:
                plot_params[k] = self.default_plot_params[k]
    else:
        plot_params = self.default_plot_params
    # if 'color' in plot_params:
    #     if type(plot_params['color']) is str:
    #         plot_params['color'] = sns.color_palette(plot_params['color'])
    #     else:
    #         plot_params['color'] = plot_params['color']
    if 'style' in plot_params:
        sns.set_style(plot_params['style']) # "ticks" "white" "dark" 'darkgrid'(default) 'whitegrid'

    if 'x_len' in plot_params:
        fig_x = plot_params['x_len']
    else:
        fig_x = 8

    'preparing the canvas'
    
    fig_collection = []
    for idx in range(len(self.data)):
        fig = plt.figure(figsize=(fig_x,5))
        if plot_params['plot_type'][0] == 'direct':
            ax = fig.add_subplot(111)
            select_subplot_type(plot_params['plot_type'][1], ax, self.data[idx], self.annotation[idx], plot_params)
        elif plot_params['plot_type'][0] == 'matrix':
            matrix_plot(fig, self.data[idx], self.annotation[idx], 
                plot_params['x_title'],plot_params['y_title'],plot_params)
        elif plot_params['plot_type'][0] == 'float':
            float_plot(fig, self.data[idx], self.annotation[idx], 
                plot_params['xy_locs'], plot_params)
        else:
            raise Exception(f'Unsupported plot_type {plot_params["plot_type"][0]}')  

        'output'
        # ipdb.set_trace()
        plt.show()
        if save:
            title = f'{title}.png'
            if type(save) == str:
                title = save
            fig.savefig(title, transparent=True)
        fig_collection.append(fig)

    'reset'
    sns.despine()
    sns.set()
    
    if return_fig:
        return fig_collection

structure.Analyzed_data.plot = plot

def select_subplot_type(subplot_type, ax, data, annotation, plot_params):
    if subplot_type == 'waveform':
        figure_unit.plot_waveform(ax, data, plot_params)
        if type(annotation) == pd.DataFrame:
            figure_unit.significant(ax, annotation, plot_params['win'], plot_params['sig_limit'])
    elif subplot_type == 'spectrum':
        figure_unit.plot_spectrum(ax, data, plot_params)
        if type(annotation) == pd.DataFrame:
            figure_unit.significant(ax, annotation, plot_params['win'], plot_params['sig_limit'])
    elif subplot_type == 'topograph':
        figure_unit.plot_topograph(ax, data, plot_params)
    elif subplot_type == 'heatmap':
        figure_unit.plot_heatmap(ax, data, plot_params)
        if type(annotation) == pd.DataFrame:
            figure_unit.heatmap_significant(ax, annotation, plot_params['sig_limit'])
    else:
        raise Exception(f'Unsupported subplot_type "{subplot_type}"')

def float_plot(fig, data, annotation, positions, plot_params):
    sns.set_style("white")

    data_cells = dict((k[2:],v) for k,v in data.groupby(level='channel_group'))

    if type(annotation) == pd.DataFrame:
        annotation_cells = dict((k[2:],v) for k,v in annotation.groupby(level='channel_group'))
    
    axs = []
    for idx,(data_name,data) in enumerate(data_cells.items()):
        data.name = data_name
        sys.stdout.write(' ' * 30 + '\r')
        sys.stdout.flush()
        sys.stdout.write('%d/%d\r' %(idx+1,len(data_cells)))
        sys.stdout.flush()

        axs.append(fig.add_axes([idx,0,0,0]))

        if type(annotation) == pd.DataFrame:
            select_subplot_type(plot_params['plot_type'][1], axs[idx], data, annotation_cells[data_name], plot_params)
        else:
            select_subplot_type(plot_params['plot_type'][1], axs[idx], data, None, plot_params)
        axs[idx].set_position([positions[data_name][0], positions[data_name][1] ,0.3,0.3])
        axs[idx].axis('off')

        axs[idx].legend(bbox_to_anchor=(1.01, 1))

    sns.set() # switch to seaborn defaults

def matrix_plot(fig, data, annotation, x_axis, y_axis, plot_params):
    sns.set_style("white")

    data = data.stack('time')

    for level_name in data.index.names:
        if '_group' in level_name:
            old_values_in_level = data.index.levels[data.index.names.index(level_name)]
            data.index = data.index.set_levels([' '.join(i.split(' ')[1:]) for i in old_values_in_level],level=level_name)
    'add "ms" to time axis'
    old_values_in_level = data.index.levels[data.index.names.index('time')]
    data.index = data.index.set_levels([f'{i}ms' for i in old_values_in_level],level='time')

    x_axis_values = list(OrderedDict.fromkeys(data.index.get_level_values(x_axis))) # remove duplicates from a list in whilst preserving order
    y_axis_values = list(OrderedDict.fromkeys(data.index.get_level_values(y_axis))) # remove duplicates from a list in whilst preserving order

    col_N,row_N = len(x_axis_values), len(y_axis_values)
    fig.set_figwidth(15)

    W,H = 1,1

    wspace = 0.1
    hspace = 0.1
    has_title = False
    has_colorbar = False

    height_ratios = [0,0.3,row_N,1]
    width_ratios = [0.7,col_N,0.1]
    fig.set_figwidth(sum(width_ratios)*1.7)
    fig.set_figheight(sum(height_ratios)*1.7)

    if has_title:
        height_ratios[0] = 0.5

    inner_grids = gridspec.GridSpec(4, 3, wspace=wspace, hspace=hspace, height_ratios=height_ratios, width_ratios=width_ratios)
    
    'title'
    ax = plt.Subplot(fig, inner_grids[1])
    ax.text(0.5, 0.5, plot_params['title'], va="center", ha="center")
    ax.axis('off')
    fig.add_subplot(ax)

    'x_axis'
    inner_grids_for_x_axis = gridspec.GridSpecFromSubplotSpec(1, col_N, inner_grids[4], wspace=wspace, hspace=hspace)
    for i in range(col_N):
        ax = plt.Subplot(fig, inner_grids_for_x_axis[i])
        ax.text(0.5, 0.5, x_axis_values[i], va="center", ha="center")
        ax.axis('off')
        fig.add_subplot(ax)

    'y_axis'
    inner_grids_for_y_axis = gridspec.GridSpecFromSubplotSpec(row_N, 1, inner_grids[6], wspace=wspace, hspace=hspace)
    for i in range(row_N):
        ax = plt.Subplot(fig, inner_grids_for_y_axis[i])
        ax.text(0.5, 0.5, y_axis_values[i], va="center", ha="center")
        ax.axis('off')
        fig.add_subplot(ax)

    'data'
    inner_grids_for_data = gridspec.GridSpecFromSubplotSpec(row_N, col_N, inner_grids[7])
    data_cells = dict((k,v) for k,v in data.groupby(level=[y_axis,x_axis]))
    for idx,data_name in enumerate(itertools.product(y_axis_values,x_axis_values)):
        ax = plt.Subplot(fig, inner_grids_for_data[idx])
        fig.add_subplot(ax)
        select_subplot_type(plot_params['plot_type'][1], ax, data_cells[data_name], annotation, plot_params)

    'color_bar'
    inner_grids_for_cbar = gridspec.GridSpecFromSubplotSpec(row_N, 1, inner_grids[8], wspace=wspace, hspace=hspace)
    ax = plt.Subplot(fig, inner_grids_for_cbar[row_N-1])
    cb1 = matplotlib.colorbar.ColorbarBase(ax, cmap=plot_params['color'] ,
                                    norm=matplotlib.colors.Normalize(vmin=plot_params['zlim'][0], vmax=plot_params['zlim'][1]),
                                    orientation='vertical')
    if 'cbar_title' in plot_params:
        ax.set_title(plot_params['cbar_title'])
    fig.add_subplot(ax)       

    sns.set() # switch to seaborn defaults



from ..default import *
import pickle

def save_epochs(epochs, filepath, append=False, all_in_one=False):
    if append:
        mode = 'a'
    else:
        mode = 'w'
    if not filepath.endswith('.h5'):
        print('Your file have been added ".h5" as the extension name.')
        filepath += '.h5'

    '''for a unknown bug of pd.HDFStore. del it here and in load.py--load_dataframe_hdf5() if fixed'''
    # trials = epochs.info['trials']
    # del epochs.info['trials']
    # timepoints = epochs.info['timepoints']['all']
    # epochs.info['timepoints']['all'] = [timepoints[0],timepoints[-1],timepoints[1]-timepoints[0]]
    ''''''

    info_t = epochs.info.copy()
    del info_t['subjects']
    del info_t['timepoints']
    del info_t['conditions']
    del info_t['channels']
    del info_t['trials']

    with pd.HDFStore(filepath, mode) as store:
        if all_in_one:
            store.put('all',epochs.all)
            store.get_storer('all').attrs['info'] = info_t
        else:
            store.put('supplementary',pd.DataFrame([0]))
            store.get_storer('supplementary').attrs['info'] = info_t
            for subj_id,subj_data in epochs.all.groupby(level=['subject']):
                subj_id = str(subj_id)
                print('saving',subj_id,'...')
                store.put(subj_id, subj_data)
                # store.get_storer(subj_id).attrs['info'] = info_t

    '''for a unknown bug of pd.HDFStore. del it here and in load.py--load_dataframe_hdf5() if fixed'''
    # epochs.info['trials'] = trials
    # epochs.info['timepoints']['all'] = timepoints
    ''''''

    print('Done.')

def save_result(result, filepath):
    if not filepath.endswith('.pickle'):
        print('Your file have been added ".pickle" as the extension name.')
        filepath += '.pickle'

    with open(filepath, 'wb') as f:
        pickle.dump(result.analysis_name, f)
        pickle.dump(result.data, f)
        pickle.dump(result.annotation, f)
        pickle.dump(result.default_plot_params, f)
        pickle.dump([i.name for i in result.data if type(i) is pd.DataFrame], f)
        pickle.dump([i.name for i in result.annotation if type(i) is pd.DataFrame], f)

    print('Result saved.')



from ..default import *
from ..structure import Epochs, Analyzed_data
import pickle
from scipy.io import loadmat

def from_mne_epochs(subj_datas,subjIDs=[],montage_path='standard-10-5-cap385'):
    '''
    Parameters:
        subj_datas: mne.epochs.Epochs or list of mne.epochs.Epochs,
    '''
    if type(subj_datas) == mne.epochs.Epochs:
        subj_datas = [subj_datas]

    if subjIDs == []:
        subjIDs = [i+1 for i in range(len(subj_datas))]

    epochs_data = []
    # load each subject's eeg data respectively 
    for subjID,epoch_raw in zip(subjIDs,subj_datas): 
        
        sr = epoch_raw.info['sfreq']
        channel_names = epoch_raw.info['ch_names']

        epochs_data_each_subj = epoch_raw.to_data_frame().stack('signal').unstack('time')

        index_new = [(subjID,condition,trial,channel) for condition,trial,channel in epochs_data_each_subj.index.tolist()]
        epochs_data_each_subj.index = pd.MultiIndex.from_tuples(index_new, names=['subject','condition','trial','channel'])

        # td = pd.to_timedelta(epochs_data_each_subj.columns,unit='ms')
        # epochs_data_each_subj.columns = [['data']*len(td), td]
        epochs_data_each_subj.columns.name = 'time'
        
        epochs_data.append(epochs_data_each_subj)
        
        print(subjID,': ')
        print('Condition count',[(k,v//len(channel_names)) for k,v in Counter(epochs_data_each_subj.index.get_level_values('condition')).items()])
        
    # combine eeg data of all the subjects
    epochs_data = pd.concat(epochs_data)    
    epochs_data.sort_index(inplace=True)

    return Epochs(epochs_data, montage_path=montage_path, info={'sample_rate':sr}   )

def load_epochs(files=[],path='',subjIDs=[],montage_path='standard-10-5-cap385',baseline_time=[],
            extremum_in_all=None,extremum_in_baseline=None,
            events_dict=dict(),drop_channels=[],compress=False):

    filepaths = infer_paths(files, path)

    if filepaths[0].endswith('.pickle'):
        return load_dataframe_pickle(files,path,subjIDs,montage_path,compress)
    elif filepaths[0].endswith('.h5'):
        return load_dataframe_hdf5(files,path,subjIDs,montage_path,compress)
    elif filepaths[0].endswith('.fif'):
        return load_mne_fif(files,path,subjIDs,montage_path)
    elif filepaths[0].endswith('.mat'):
        return load_eeglab_mat(files,path,subjIDs,montage_path,baseline_time,
            extremum_in_all,extremum_in_baseline,events_dict,drop_channels)
    else:
        raise ValueError('Only supported pickle files (".pickle"), HDF5 files (".h5"), MNE epochs files(".fif"), or EEGLAB files (".mat")')

def infer_paths(files, path):
    if type(files)==str:
        files = [files]
    
    if len(files)==0:
        if path!='':
            filepaths = os.listdir(path)
        else:
            filepaths = os.listdir('data')
    elif len(files)>0:
        if path!='':
            filepaths = [path+'/'+fn for fn in files]
        else:
            filepaths = files

    return filepaths

def load_dataframe_pickle(files=[],path='',subjIDs=[],montage_path='standard-10-5-cap385',compress=False):
    filepaths = infer_paths(files, path)

    sys.modules['pandas.tseries.tdi'] = pd
    sys.modules['pandas.indexes'] = pd.core.indexes

    epochs_data = []
    for filepath in filepaths:
        try:
            df = pd.read_pickle(filepath)
            if type(df) is pd.DataFrame:
                if compress:
                    conditions_list = list(df.index.get_level_values('condition').unique())
                    df_averaged = df.mean(level=list(np.setdiff1d(df.index.names, ['trial'])))
                    df_averaged.index = pd.MultiIndex.from_tuples(
                                        [(subject,condition,conditions_list.index(condition),channel) 
                                        for (channel,condition,subject) in df_averaged.index],
                                        names=['subject','condition','trial','channel'])
                    epochs_data.append(df_averaged)
                else:
                    epochs_data.append(df)
            else:
                print('"%s" is not a DataFrame file' %filename)
        except:
            print('"%s" cannot be loaded as pickle file' %filename)

    epochs_data = pd.concat(epochs_data)
    epochs_data.sort_index(inplace=True)

    try:
        epochs_data.columns = [int(i.to_timedelta64())//1000000 for i in epochs_data['data'].columns]
    except:
        pass

    return Epochs(epochs_data, montage_path=montage_path)

def load_dataframe_hdf5(files=[],path='',subjIDs=[],montage_path='standard-10-5-cap385',compress=False):
    def preprocess(df):
        if type(df) is pd.DataFrame:
            if compress:
                conditions_list = list(df.index.get_level_values('condition').unique())
                df_averaged = df.mean(level=list(np.setdiff1d(df.index.names, ['trial'])))
                df_averaged.index = pd.MultiIndex.from_tuples(
                                    [(subject,condition,conditions_list.index(condition),channel) 
                                    for (channel,condition,subject) in df_averaged.index],
                                    names=['subject','condition','trial','channel'])
                return df_averaged
            else:
                return df
        else:
            raise Exception(f'{sub_id} in "{filepath}" is not a DataFrame')

    filepaths = infer_paths(files, path)

    epochs_data = []
    info = dict()
    for filepath in filepaths:
        try:
            store = pd.HDFStore(filepath)
            if list(store.keys())==['/all']:
                df = store['/all']
                info = store.get_storer('/all').attrs['info']
                epochs_data.append(preprocess(df))
            for sub_id in store.keys():
                print(f'loading subject{sub_id}')
                df = store[sub_id]
                info = store.get_storer(sub_id).attrs['info']
                epochs_data.append(preprocess(df))
            store.close()
        except:
            raise Exception('"%s" cannot be loaded as a hdf5 file' %filepath)

    if len(epochs_data)>1:
        epochs_data = pd.concat(epochs_data)

    epochs_data.sort_index(inplace=True)

    '''for a unknown bug of pd.HDFStore. del it here and in save.py if fixed'''
    info['trials'] = dict()
    info['trials']['all'] = list(epochs_data.index.get_level_values('trial').unique())
    for subj_id,subj_data in epochs_data.groupby(level=['subject']):
        info['trials'][str(subj_id)] = sorted(list(subj_data.index.get_level_values('trial').unique()))
    info['timepoints'] = {'all':list(epochs_data.columns)}
    '''for a unknown bug of pd.HDFStore'''

    return Epochs(epochs_data, montage_path=montage_path, info=info)

def load_dataframe_hdf5(files=[],path='',subjIDs=[],montage_path='standard-10-5-cap385',compress=False):
    def preprocess(df):
        if type(df) is pd.DataFrame:
            if compress:
                conditions_list = list(df.index.get_level_values('condition').unique())
                df_averaged = df.mean(level=list(np.setdiff1d(df.index.names, ['trial'])))
                df_averaged.index = pd.MultiIndex.from_tuples(
                                    [(subject,condition,conditions_list.index(condition),channel) 
                                    for (channel,condition,subject) in df_averaged.index],
                                    names=['subject','condition','trial','channel'])
                return df_averaged
            else:
                return df
        else:
            raise Exception(f'{subjID} in "{filepath}" is not a DataFrame')

    filepaths = infer_paths(files, path)

    epochs_data = []
    info = dict()
    for filepath in filepaths:
        try:
            store = pd.HDFStore(filepath)
            if list(store.keys())==['/all']:
                df = store['/all']
                info = store.get_storer('/all').attrs['info']
                epochs_data.append(preprocess(df))
                print
            if subjIDs==[]:
                subjIDs = store.keys()
            for subjID in subjIDs:
                if subjID != '/supplementary':
                    print(f'reading {subjID[1:]}')
                    df = store[subjID]
                    epochs_data.append(preprocess(df))
                else:
                    info = store.get_storer('/supplementary').attrs['info']
            store.close()
        except Exception as inst:
            raise Exception(f'{inst}. So "{filepath}" cannot be loaded as a hdf5 file.')

    if len(epochs_data)>1:
        epochs_data = pd.concat(epochs_data)
        print('Concatenated.')

    epochs_data.sort_index(inplace=True)

    return Epochs(epochs_data, montage_path=montage_path, info=info)

def load_mne_fif(files=[],path='',subjIDs=[],montage_path='standard-10-5-cap385'):
    '''
    Parameters:
        paths: FIF filepath or list of FIF filepath
    '''
    filepaths = infer_paths(files, path)

    if type(filepaths[0]) is str:
        subj_datas = [mne.read_epochs(path) for path in filepaths]
    else:
        raise ValueError('Parameter [files] should be path(s)')

    return from_mne_epochs(subj_datas,subjIDs=[],montage_path='standard-10-5-cap385')

def load_eeglab_mat(files=[],path='',subjIDs=[],montage_path='standard-10-5-cap385',baseline_time=[],
            extremum_in_all=None,extremum_in_baseline=None,
            events_dict=dict(),drop_channels=[]):

    filepaths = infer_paths(files, path)

    if subjIDs == []:
        subjIDs = [i+1 for i in range(len(filepaths))]

    epochs_data = []
    # load each subject's eeg data respectively 
    for subjID,filepath in zip(subjIDs,filepaths): 
        # load .mat file
        mat_file = loadmat(filepath)['EEG'][0][0]
        ## get information in mat
        # chanlocs_raw = mat_file[21][0]
        # chanlocs = dict()
        # for l in chanlocs_raw:
        #     chan_name = l[0][0]
        #     # if chan_name in drop_channels:
        #     #     continue
        #     try:
        #         theta = l[2][0][0]
        #         r = l[3][0][0]
        #         chanlocs[chan_name] = [r*math.sin(np.pi/180*theta),r*math.cos(np.pi/180*theta)]
        #     except:
        #         chanlocs[chan_name] = None
        # parameter.xy_locs = chanlocs
        channel_names = [c[0][0] for c in mat_file[21][0]]

        eeg_data = mat_file[15]
        sr = mat_file[11][0][0]
        epoch_time = (int(mat_file[12]*1000),int(mat_file[13]*1000))

        trial_count = mat_file[15].shape[2]
        events = [i[4][0] for i in mat_file[25][0]]
        if events_dict==dict():
            events_dict = {m:m for m in set(events)}

        channels = dict() # build up a dictionary contained id2channel and channel2id
        for ind,i in enumerate(mat_file[21][0]):
            channels[i[0][0]] = ind
            channels[ind] = i[0][0]
        channel_count = mat_file[15].shape[0]
        channel_names = [i[0][0] for i in mat_file[21][0]]

        # find the indexs of timepoints
        # epoch_start = (epoch_time[0]-raw_epoch_time[0])*sr//1000
        # epoch_end = (epoch_time[1]-raw_epoch_time[0])*sr//1000
        if baseline_time != []:
            baseline_start = (baseline_time[0]-epoch_time[0])*sr//1000
            baseline_end = (baseline_time[1]-epoch_time[1])*sr//1000
            baseline_end_new = -baseline_time[0]*sr//1000
        
        # add eeg data
        count = 0
        epochs_data_each_subj = []
        indexs_each_subj = []
        for T in range(trial_count):
            reject = False
            epochs_data_each_trial = []
            indexs_each_trial = []
            
            for C in range(channel_count):
                eeg = eeg_data[C,:,T]

                if baseline_time != []:
                    eeg -= eeg[baseline_start:baseline_end].mean()

                if extremum_in_all:
                    keep1 = min(eeg)> -extremum_in_all and max(eeg) < extremum_in_all
                else:
                    keep1 = True
                if extremum_in_baseline:
                    keep2 = max(eeg[:baseline_end_new])-min(eeg[:baseline_end_new]) < extremum_in_baseline
                else:
                    keep2 = True

                if keep1 and keep2:
                    epochs_data_each_trial.append(np.array(eeg))
                    indexs_each_trial.append([subjID,events_dict[events[T]],T,channels[C]])
                else:
                    reject = True
                    
            if not reject:
                epochs_data_each_subj += epochs_data_each_trial
                indexs_each_subj += indexs_each_trial
                count+=1

        # convert the data to dataframe
        # set up a timeseries for dataframe
        timespan = np.arange(epoch_time[0], epoch_time[1]+1000//sr, 1000//sr)
        index = pd.MultiIndex.from_tuples(indexs_each_subj, names=['subject','condition','trial','channel'])
        epochs_data_each_subj = pd.DataFrame(epochs_data_each_subj, index=index, columns = timespan)
        epochs_data_each_subj.columns.name = 'time'

        # set up a timeseries index for eeg data. eg. start='-200 ms', end='799 ms', freq='2ms'
        # span = pd.timedelta_range(start='%d ms' %epoch_time[0], end='%d ms' %(epoch_time[1]), freq='%d ms' %(1000/sr)) 
        # epochs_data_each_subj.columns = pd.MultiIndex.from_tuples([('data', ts) for ts in span], names=['','time'])

        epochs_data.append(epochs_data_each_subj)
        print(subjID,': ',round(count/trial_count,2),' in ',trial_count)
        print('Condition count',[(k,v//len(channel_names)) for k,v in Counter(epochs_data_each_subj.index.get_level_values('condition')).items()])

    # combine eeg data of all the subjects
    epochs_data = pd.concat(epochs_data)    
    epochs_data.sort_index(inplace=True)

    return Epochs(epochs_data, montage_path=montage_path, info={'sample_rate':sr})

def load_raw_eeg(files=[],path='',subjIDs=[],montage_path='standard-10-5-cap385',events_dict=dict(),default_reference_channels='Cz',drop_channels=[],filter_range=(0.1, 30.),ref_channels='average',epoch_range=(-0.2, 1) ,reject=None):
    filepaths = infer_paths(files, path)

    if type(events_dict) is list:
            events_dict = {k:k for k in events_dict}

    if subjIDs == []:
        subjIDs = [i+1 for i in range(len(filepaths))]

    epochs_data = []
    # load each subject's eeg data respectively 
    for filepath in filepaths:
        print(filepath)
        if  filepath[-4:] in ['.edf','.bdf','gdf']:
            read_eeg_file = mne.io.read_raw_edf
        elif filepath[-5:] == '.vhdr':
            read_eeg_file = mne.io.read_raw_brainvision
        elif filepath[-5:] == '.fif':
            read_eeg_file = mne.io.read_raw_fif
        elif filepath[-4:] in ['.egi','.mff']:
            read_eeg_file = mne.io.read_raw_egi
        else:
            raise Exception(f'{filepath} is an unsupported file type')

        EEG = read_eeg_file(filepath,event_id=events_dict, preload=True)

        EEG = mne.add_reference_channels(EEG, default_reference_channels)
        EEG.drop_channels(drop_channels)

        EEG.set_montage(mne.channels.read_montage('standard_1020'))

        EEG.filter(filter_range[0], filter_range[1])

        EEG.set_eeg_reference(ref_channels=ref_channels)  # set EEG average reference
        EEG.apply_proj() # Average reference projection was added, but hasn't been applied yet. Use the .apply_proj() method function to apply projections.

        epoch = mne.Epochs(EEG, EEG._events, events_dict, epoch_range[0], epoch_range[1], baseline=(epoch_range[0], 0), preload=True, reject=reject)
        epochs_data.append(epoch)

    epochs_data = from_mne_epochs(epochs_data,subjIDs=subjIDs,montage_path=montage_path)
    return epochs_data

def load_filetrip_mat():
    pass

def load_AnalyzedData(filepath):
    with open(filepath, 'rb') as f:
        analysis_name = pickle.load(f)
        data = pickle.load(f)
        annotation = pickle.load(f)
        default_plot_params = pickle.load(f)
        data_names = pickle.load(f)
        annotation_names = pickle.load(f)

        for idx,i in enumerate(data_names):
            data[idx].name = i
        for idx,i in enumerate(annotation_names):
            annotation[idx].name = i

        return Analyzed_data(analysis_name, data, annotation, default_plot_params)

'copied a lot from MNE 0.14.1'
def load_topolocs(f_path,ch_names):
    def read_montage(kind, ch_names=None, path=None, unit='m', transform=False):
        """Read a generic (built-in) montage.

        Individualized (digitized) electrode positions should be
        read in using :func:`read_dig_montage`.

        In most cases, you should only need the `kind` parameter to load one of
        the built-in montages (see Notes).

        Parameters
        ----------
        kind : str
            The name of the montage file without the file extension (e.g.
            kind='easycap-M10' for 'easycap-M10.txt'). Files with extensions
            '.elc', '.txt', '.csd', '.elp', '.hpts', '.sfp' or '.loc' ('.locs' and
            '.eloc') are supported.
        ch_names : list of str | None
            If not all electrodes defined in the montage are present in the EEG
            data, use this parameter to select subset of electrode positions to
            load. If None (default), all defined electrode positions are returned.

            .. note:: ``ch_names`` are compared to channel names in the montage
                      file after converting them both to upper case. If a match is
                      found, the letter case in the original ``ch_names`` is used
                      in the returned montage.

        path : str | None
            The path of the folder containing the montage file. Defaults to the
            mne/channels/data/montages folder in your mne-python installation.
        unit : 'm' | 'cm' | 'mm'
            Unit of the input file. If not 'm' (default), coordinates will be
            rescaled to 'm'.
        transform : bool
            If True, points will be transformed to Neuromag space.
            The fidicuals, 'nasion', 'lpa', 'rpa' must be specified in
            the montage file. Useful for points captured using Polhemus FastSCAN.
            Default is False.

        Returns
        -------
        pos, ch_names_, kind, selection

        See Also
        --------
        DigMontage
        Montage
        read_dig_montage

        Notes
        -----
        Built-in montages are not scaled or transformed by default.

        Montages can contain fiducial points in addition to electrode
        locations, e.g. ``biosemi64`` contains 67 total channels.

        The valid ``kind`` arguments are:

        ===================   =====================================================
        Kind                  description
        ===================   =====================================================
        standard_1005         Electrodes are named and positioned according to the
                              international 10-05 system.
        standard_1020         Electrodes are named and positioned according to the
                              international 10-20 system.
        standard_alphabetic   Electrodes are named with LETTER-NUMBER combinations
                              (A1, B2, F4, etc.)
        standard_postfixed    Electrodes are named according to the international
                              10-20 system using postfixes for intermediate
                              positions.
        standard_prefixed     Electrodes are named according to the international
                              10-20 system using prefixes for intermediate
                              positions.
        standard_primed       Electrodes are named according to the international
                              10-20 system using prime marks (' and '') for
                              intermediate positions.

        biosemi16             BioSemi cap with 16 electrodes
        biosemi32             BioSemi cap with 32 electrodes
        biosemi64             BioSemi cap with 64 electrodes
        biosemi128            BioSemi cap with 128 electrodes
        biosemi160            BioSemi cap with 160 electrodes
        biosemi256            BioSemi cap with 256 electrodes

        easycap-M10           Brainproducts EasyCap with electrodes named
                              according to the 10-05 system
        easycap-M1            Brainproduct EasyCap with numbered electrodes

        EGI_256               Geodesic Sensor Net with 256 channels

        GSN-HydroCel-32       HydroCel Geodesic Sensor Net with 32 electrodes
        GSN-HydroCel-64_1.0   HydroCel Geodesic Sensor Net with 64 electrodes
        GSN-HydroCel-65_1.0   HydroCel Geodesic Sensor Net with 64 electrodes + Cz
        GSN-HydroCel-128      HydroCel Geodesic Sensor Net with 128 electrodes
        GSN-HydroCel-129      HydroCel Geodesic Sensor Net with 128 electrodes + Cz
        GSN-HydroCel-256      HydroCel Geodesic Sensor Net with 256 electrodes
        GSN-HydroCel-257      HydroCel Geodesic Sensor Net with 256 electrodes + Cz
        ===================   =====================================================

        .. versionadded:: 0.9.0
        """
        if path is None:
            path = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'data', 'montages')
            supported = ('.elc', '.txt', '.csd', '.sfp', '.elp', '.hpts', '.loc',
                         '.locs', '.eloc')
            montages = [os.path.splitext(f) for f in os.listdir(path)]
            montages = [m for m in montages if m[1] in supported and kind == m[0]]
            if len(montages) != 1:
                raise ValueError('Could not find the montage. Please provide the '
                                 'full path.')
            kind, ext = montages[0]
        else:
            kind, ext = os.path.splitext(kind)
        fname = os.path.join(path, kind + ext)

        if ext == '.sfp':
            # EGI geodesic
            with open(fname, 'r') as f:
                lines = f.read().replace('\t', ' ').splitlines()

            ch_names_, pos = [], []
            for ii, line in enumerate(lines):
                line = line.strip().split()
                if len(line) > 0:  # skip empty lines
                    if len(line) != 4:  # name, x, y, z
                        raise ValueError("Malformed .sfp file in line " + str(ii))
                    this_name, x, y, z = line
                    ch_names_.append(this_name)
                    pos.append([float(cord) for cord in (x, y, z)])
            pos = np.asarray(pos)
        elif ext == '.elc':
            # 10-5 system
            ch_names_ = []
            pos = []
            with open(fname) as fid:
                # Default units are meters
                for line in fid:
                    if 'UnitPosition' in line:
                        units = line.split()[1]
                        scale_factor = dict(m=1., mm=1e-3)[units]
                        break
                else:
                    raise RuntimeError('Could not detect units in file %s' % fname)
                for line in fid:
                    if 'Positions\n' in line:
                        break
                pos = []
                for line in fid:
                    if 'Labels\n' in line:
                        break
                    pos.append(list(map(float, line.split())))
                for line in fid:
                    if not line or not set(line) - set([' ']):
                        break
                    ch_names_.append(line.strip(' ').strip('\n'))
            pos = np.array(pos) * scale_factor
        elif ext == '.txt':
            # easycap
            try:  # newer version
                data = np.genfromtxt(fname, dtype='str', skip_header=1)
            except TypeError:
                data = np.genfromtxt(fname, dtype='str', skiprows=1)
            ch_names_ = list(data[:, 0])
            az = np.deg2rad(data[:, 2].astype(float))
            pol = np.deg2rad(data[:, 1].astype(float))
            pos = _sph_to_cart(np.array([np.ones(len(az)) * 85., az, pol]).T)
        elif ext == '.csd':
            # CSD toolbox
            dtype = [('label', 'S4'), ('theta', 'f8'), ('phi', 'f8'),
                     ('radius', 'f8'), ('x', 'f8'), ('y', 'f8'), ('z', 'f8'),
                     ('off_sph', 'f8')]
            try:  # newer version
                table = np.loadtxt(fname, skip_header=2, dtype=dtype)
            except TypeError:
                table = np.loadtxt(fname, skiprows=2, dtype=dtype)
            ch_names_ = table['label']
            az = np.deg2rad(table['theta'])
            pol = np.deg2rad(90. - table['phi'])
            pos = _sph_to_cart(np.array([np.ones(len(az)), az, pol]).T)
        elif ext == '.elp':
            # standard BESA spherical
            dtype = np.dtype('S8, S8, f8, f8, f8')
            try:
                data = np.loadtxt(fname, dtype=dtype, skip_header=1)
            except TypeError:
                data = np.loadtxt(fname, dtype=dtype, skiprows=1)

            ch_names_ = data['f1'].astype(np.str)
            az = data['f2']
            horiz = data['f3']
            radius = np.abs(az / 180.)
            az = np.deg2rad(np.array([h if a >= 0. else 180 + h
                                      for h, a in zip(horiz, az)]))
            pol = radius * np.pi
            pos = _sph_to_cart(np.array([np.ones(len(az)) * 85., az, pol]).T)
        elif ext == '.hpts':
            # MNE-C specified format for generic digitizer data
            dtype = [('type', 'S8'), ('name', 'S8'),
                     ('x', 'f8'), ('y', 'f8'), ('z', 'f8')]
            data = np.loadtxt(fname, dtype=dtype)
            ch_names_ = data['name'].astype(np.str)
            pos = np.vstack((data['x'], data['y'], data['z'])).T
        elif ext in ('.loc', '.locs', '.eloc'):
            ch_names_ = np.loadtxt(fname, dtype='S4',
                                   usecols=[3]).astype(np.str).tolist()
            dtype = {'names': ('angle', 'radius'), 'formats': ('f4', 'f4')}
            topo = np.loadtxt(fname, dtype=float, usecols=[1, 2])
            sph = _topo_to_sph(topo)
            pos = _sph_to_cart(sph)
            pos[:, [0, 1]] = pos[:, [1, 0]] * [-1, 1]
        else:
            raise ValueError('Currently the "%s" template is not supported.' %
                             kind)
        selection = np.arange(len(pos))

        if unit == 'mm':
            pos /= 1e3
        elif unit == 'cm':
            pos /= 1e2
        elif unit != 'm':
            raise ValueError("'unit' should be either 'm', 'cm', or 'mm'.")
        if transform:
            names_lower = [name.lower() for name in list(ch_names_)]
            if ext == '.hpts':
                fids = ('2', '1', '3')  # Alternate cardinal point names
            else:
                fids = ('nasion', 'lpa', 'rpa')

            missing = [name for name in fids
                       if name not in names_lower]
            if missing:
                raise ValueError("The points %s are missing, but are needed "
                                 "to transform the points to the MNE coordinate "
                                 "system. Either add the points, or read the "
                                 "montage with transform=False. " % missing)
            nasion = pos[names_lower.index(fids[0])]
            lpa = pos[names_lower.index(fids[1])]
            rpa = pos[names_lower.index(fids[2])]

            neuromag_trans = get_ras_to_neuromag_trans(nasion, lpa, rpa)
            pos = apply_trans(neuromag_trans, pos)

        if ch_names is not None:
            # Ensure channels with differing case are found.
            upper_names = [ch_name.upper() for ch_name in ch_names]
            sel, ch_names_ = zip(*[(i, ch_names[upper_names.index(e)]) for i, e in
                                   enumerate([n.upper() for n in ch_names_])
                                   if e in upper_names])
            sel = list(sel)
            pos = pos[sel]
            selection = selection[sel]
        else:
            ch_names_ = list(ch_names_)
        kind = os.path.split(kind)[-1]
        return pos, ch_names_, kind, selection
    
    def _sph_to_cart(sph):
        """Convert spherical coordinates to Cartesion coordinates."""
        assert sph.ndim == 2 and sph.shape[1] == 3
        sph = np.atleast_2d(sph)
        out = np.empty((len(sph), 3))
        out[:, 2] = sph[:, 0] * np.cos(sph[:, 2])
        xy = sph[:, 0] * np.sin(sph[:, 2])
        out[:, 0] = xy * np.cos(sph[:, 1])
        out[:, 1] = xy * np.sin(sph[:, 1])
        return out

    def _cart_to_sph(cart):
        """Convert Cartesian coordinates to spherical coordinates.

        Parameters
        ----------
        cart_pts : ndarray, shape (n_points, 3)
            Array containing points in Cartesian coordinates (x, y, z)

        Returns
        -------
        sph_pts : ndarray, shape (n_points, 3)
            Array containing points in spherical coordinates (rad, azimuth, polar)
        """
        assert cart.ndim == 2 and cart.shape[1] == 3
        cart = np.atleast_2d(cart)
        out = np.empty((len(cart), 3))
        out[:, 0] = np.sqrt(np.sum(cart * cart, axis=1))
        out[:, 1] = np.arctan2(cart[:, 1], cart[:, 0])
        out[:, 2] = np.arccos(cart[:, 2] / out[:, 0])
        out = np.nan_to_num(out)
        return out

    def _pol_to_cart(pol):
        """Transform polar coordinates to cartesian."""
        out = np.empty((len(pol), 2))
        if pol.shape[1] == 2:  # phi, theta
            out[:, 0] = pol[:, 0] * np.cos(pol[:, 1])
            out[:, 1] = pol[:, 0] * np.sin(pol[:, 1])
        else:  # radial distance, theta, phi
            d = pol[:, 0] * np.sin(pol[:, 2])
            out[:, 0] = d * np.cos(pol[:, 1])
            out[:, 1] = d * np.sin(pol[:, 1])
        return out
    
    pos, ch_names, kind, selection = read_montage(f_path,ch_names)
    new_pos = _pol_to_cart(_cart_to_sph(pos)[:, 1:][:, ::-1])
    
    xy_locs = dict([(ch,[x,y]) for ch,(x,y) in zip(ch_names,new_pos)])
    return xy_locs



import numpy as np
import scipy
import pandas as pd
import seaborn as sns
import matplotlib
import matplotlib.pyplot as plt

import mne

# import scipy.io

from scipy import stats
import statsmodels
import statsmodels.api as sm
from statsmodels.formula.api import ols
import permute
from permute.core import two_sample

import random
# from random import sample
import math
import collections
from collections import defaultdict
from collections import OrderedDict
from collections import Counter
from collections import namedtuple
import itertools
import re
import time
import sys
import os
import warnings

import ipdb


ids = pd.IndexSlice
td = pd.Timedelta

warnings.filterwarnings("ignore")

tableau20 = [(31, 119, 180), (174, 199, 232), (255, 127, 14), (255, 187, 120),  
             (44, 160, 44), (152, 223, 138), (214, 39, 40), (255, 152, 150),  
             (148, 103, 189), (197, 176, 213), (140, 86, 75), (196, 156, 148),  
             (227, 119, 194), (247, 182, 210), (127, 127, 127), (199, 199, 199),  
             (188, 189, 34), (219, 219, 141), (23, 190, 207), (158, 218, 229)]  

tableau10 = [ (31, 119, 180),(255, 127, 14), (44, 160, 44),(214, 39, 40),
              (148, 103, 189), (140, 86, 75),
             (127, 127, 127), (23, 190, 207), (188, 189, 34), (227, 119, 194)]

# Scale the RGB values to the [0, 1] range, which is the format matplotlib accepts.  
for i in range(len(tableau20)):  
    r, g, b = tableau20[i]  
    tableau20[i] = (r / 255., g / 255., b / 255.) 
    
# Scale the RGB values to the [0, 1] range, which is the format matplotlib accepts.  
for i in range(len(tableau10)):  
    r, g, b = tableau10[i]  
    tableau10[i] = (r / 255., g / 255., b / 255.) 


from ..default import *
from .. import structure
from .basic import * 
from ..statistics import stats_methods

import sklearn
from sklearn import linear_model
from sklearn.model_selection import StratifiedShuffleSplit
from sklearn.metrics import roc_auc_score

def reshape_X(X, extra_params):
    X = np.array([np.array(x)[0] for x in X])
    return X

def reshape_Y(Y, extra_params):
    return sklearn.preprocessing.LabelBinarizer().fit_transform(Y)

def sklearn_model(X, Y, train_index, test_index, model):
    model.fit(X[train_index], Y[train_index])

    if getattr(model,'predict_proba', None):
        prob_train = model.predict_proba(X[train_index])[:,1]
        prob_test = model.predict_proba(X[test_index])[:,1]
    else:
        prob_train = model.decision_function(X[train_index])
        prob_test = model.decision_function(X[test_index])

    score_train = roc_auc_score(Y[train_index],prob_train)
    score_test = roc_auc_score(Y[test_index],prob_test)
    # coef = model.coef_[0] # weights
    return score_train, score_test

def group_stats(df, reshape_X_method, reshape_Y_method, run_model, fold, test_size, extra_params=dict()):
    result_real = []
    baselines = []
    for subject_group_id,subject_group_data in df.groupby(level='subject'):  # df: channel,condition_group,trial
        X, Y = zip(*[(data.T, name[0]) for name,data in subject_group_data.groupby(level=['condition_group','trial'])])
        X, Y = reshape_X_method(X, extra_params), reshape_Y_method(Y, extra_params)

        score_train_folds = []
        score_test_folds = []
        for train_index, test_index in StratifiedShuffleSplit(fold, test_size=test_size).split(X, Y):
            import types
            if type(run_model) is types.FunctionType:
                score_train, score_test = run_model(X, Y, train_index, test_index, extra_params)
            else:
                score_train, score_test = sklearn_model(X, Y, train_index, test_index, run_model)

            score_train_folds.append(score_train)
            score_test_folds.append(score_test)

        result_real.append(np.mean(score_test_folds))
        baselines.append(1/2)

    # t,pvalue = scipy.stats.ttest_rel(result_real, baselines)
    pvalue,t = two_sample(result_real, baselines, reps=1000,stat=lambda u,v: np.mean(u-v),alternative='two-sided')
    return pvalue, np.mean(result_real)

def classification(self, step_size='1ms', win_size='1ms', sample='mean', run_model=linear_model.LogisticRegression(class_weight="balanced"), reshape_X_method=reshape_X, reshape_Y_method=reshape_Y, fold=30, test_size=0.3, extra_params=dict(), parallel=False):
    # with the decorator, we can just focuse on case data instead of batch/collection data
    @self.iter('all')
    def learn(case_raw_data):
        case_raw_data = sampling(case_raw_data, step_size, win_size, sample)
        check_availability(case_raw_data, 'condition_group', '==2') 

        return roll_on_levels(case_raw_data, group_stats, arguments_dict=dict(run_model=run_model, reshape_X_method=reshape_X_method, reshape_Y_method=reshape_Y_method, fold=fold, test_size=test_size, extra_params=extra_params), levels='time', prograssbar=True, parallel=parallel)

    learning_collection, annotation_collection = learn()
    default_plot_params = dict(title='Pattern classification',plot_type=['direct','heatmap'], x_len=12, re_assign=[(0,0.01,0.05,1),(3,2,1)],
                                color=sns.cubehelix_palette(light=1, as_cmap=True), grid=True, 
                                x_title='time', y_title='condition_group',cbar_title='pvalue',cbar_values=['>=0.05','<0.05','<0.01'])

    return structure.Analyzed_data('Pattern classification', learning_collection, annotation_collection, default_plot_params=default_plot_params)



from ..default import *
from .. import structure
from .basic import * 
from ..statistics import stats_methods

from scipy.spatial.distance import cosine
from sklearn.cluster import KMeans
        
def clustering(self, n_clusters):
    def relabel(labels,centers):
        dis = [1-cosine(centers[labels[0]],centers[j]) for j in range(n_clusters)]
        mapping = dict(zip(sorted(range(len(dis)), key=lambda k: dis[k],reverse = True),range(n_clusters)))

        new_labels = [mapping[i] for i in labels]
        new_centers = {mapping[ind]:center for ind,center in enumerate(centers)}
        return new_labels,new_centers

    def kmeans(data):
        condition_order = data.index.get_level_values('condition_group').unique()
        kmeans = KMeans(n_clusters=n_clusters)
        kmeans.fit(data)
        labels = kmeans.predict(data)
        centers = kmeans.cluster_centers_.squeeze()

        new_labels,new_centers = relabel(labels,centers)

        clusterID = pd.Series(new_labels,index=data.index).unstack('time')
        clusterID = clusterID.reindex(condition_order) # 使condition的顺序和定义时一致
        clusterID.name = data.name
        return clusterID

    # with the decorator, we can just focuse on case data instead of batch/collection data
    @self.iter('average')
    def to_erp(case_raw_data):
        return case_raw_data.mean(level=['condition_group','channel']).stack('time').unstack('channel')

    erp_collection = to_erp()

    clustering_data = [kmeans(erp_batch) for erp_batch in erp_collection]

    default_plot_params = dict(plot_type=['direct','heatmap'], color=sns.cubehelix_palette(light=0.95,as_cmap=True), cbar_values=list(range(1,n_clusters+1)),cbar_title='clusters',
        grid=True, style='darkgrid')
    return structure.Analyzed_data('Clustering', clustering_data, None, default_plot_params)



from .default import *

from . import io
from . import group
from . import structure

from . import graph
from . import statistics
from . import algorithms

import importlib
def reload(module=None):
    for _ in range(2):
        importlib.reload(io.save)
        importlib.reload(io.load)
        importlib.reload(io)
        importlib.reload(structure)
        importlib.reload(structure.group)
        importlib.reload(group)
        importlib.reload(algorithms)
        if module != None: 
            importlib.reload(module)  
        importlib.reload(algorithms.basic)
        importlib.reload(algorithms.stats_methods) 
        importlib.reload(graph.figure_group) 
        importlib.reload(graph.figure_unit)

print('EasyEEG loaded.')


from ..default import *
from .. import structure
from .basic import * 
from ..statistics import stats_methods
from scipy import signal

def topography(self, step_size='1ms', win_size='1ms', sample='mean', sig_limit=0.05):
    @self.iter('average')
    def to_topo(case_raw_data):
        check_availability(case_raw_data, 'channel', '>1')

        data_with_subject = case_raw_data.mean(
            level=['subject', 'condition_group', 'channel'])

        data_with_subject = sampling(
            data_with_subject, step_size, win_size, sample)

        if len(data_with_subject.index.get_level_values('condition_group').unique()) == 2:  
            condition_groups_data = [data for idx,data in data_with_subject[0].groupby(level='condition_group')]
            
            topo_result = condition_groups_data[0].mean(level='channel') - condition_groups_data[1].mean(level='channel')
            recover_index(topo_result, data_with_subject,'condition_group') # re-add level 'condition_group' in index
            stats_result = roll_on_levels_and_compare(data_with_subject, stats_methods.t_test, levels=['time','channel'], 
                              between='condition_group', in_group='subject',prograssbar=False)
            # [0] is ugly
            result = pd.concat([topo_result, stats_result[0]],
                               keys=['Amp', 'p_val'], axis=1)

        else:
            result = data_with_subject.mean(level=['condition_group','channel'])
            result.columns = pd.MultiIndex.from_product([['Amp'], result[0].columns],names=['','time'])

        return result

    topo_collection = to_topo()

    minmax = [(t.min().min(),t.max().max()) for t in topo_collection]
    minmax = (np.array(minmax).min(),np.array(minmax).max())

    default_plot_params = dict(title='Topography',plot_type=['matrix','topograph'], zlim=minmax, color=plt.cm.jet, cbar_title='uV',
        chan_locs=self.info['xy_locs'], sig_limit=sig_limit, x_title='time', y_title='condition_group')
    return structure.Analyzed_data('Topography', topo_collection, default_plot_params=default_plot_params)

def frequency_topography(self, step_size='1ms', win_size='1ms', sample='mean', sig_limit=0.05, target=10):
    if isinstance(target, (int, float)):
        freqs = [target]
    elif type(target) is list and len(target) == 2 \
            and isinstance(target[0], (int, float)) and isinstance(target[1], (int, float)):
        freqs = np.arange(target[0], target[1])
    else:
        raise Exception(
            'Unsupported value for "target". The value should be a number, or a list of two numbers')
    
    if freqs[0] == 0: freqs += 0.001
    
    def cwt(data):
        cwt_result = signal.cwt(
            np.array(data)[0], signal.ricker, widths=freqs).mean(axis=0)
        cwt_result = pd.DataFrame(
            [cwt_result], index=data.index, columns=data.columns)
        return cwt_result

    @self.iter('average')
    def to_topo(case_raw_data):
        data_without_subject = convert(
            case_raw_data, ['condition_group', 'channel'], cwt)
        data_without_subject = sampling(data_without_subject, step_size, win_size, sample)

        if len(data_without_subject.index.get_level_values('condition_group').unique()) == 2:
            data_with_subject = convert(
                case_raw_data, ['subject', 'condition_group', 'channel'], cwt)
            data_with_subject = sampling(
                data_with_subject, step_size, win_size, sample)

            condition_groups_data = [
                data for idx, data in data_with_subject[0].groupby(level='condition_group')]

            topo_result = condition_groups_data[0].mean(
                level='channel') - condition_groups_data[1].mean(level='channel')
            # re-add level 'condition_group' in index
            recover_index(topo_result, data_with_subject, 'condition_group')
            stats_result = roll_on_levels_and_compare(data_with_subject, stats_methods.t_test, levels=['time', 'channel'],
                                                      between='condition_group', in_group='subject', prograssbar=False)
            # [0] is ugly
            result = pd.concat([topo_result, stats_result[0]],
                               keys=['Amp', 'p_val'], axis=1)

        else:
            result = data_without_subject
            result.columns = pd.MultiIndex.from_product(
                [['Amp'], result[0].columns], names=['', 'time'])

        return result

    topo_collection = to_topo()

    minmax = [(t.min().min(), t.max().max()) for t in topo_collection]
    minmax = (np.array(minmax).min(), np.array(minmax).max())

    default_plot_params = dict(title='Topography', plot_type=['matrix', 'topograph'], zlim=minmax, color=plt.cm.jet, cbar_title='Power',
                               chan_locs=self.info['xy_locs'], sig_limit=sig_limit, x_title='time', y_title='condition_group')
    return structure.Analyzed_data('Topography', topo_collection, default_plot_params=default_plot_params)

def significant_channels_count(self, step_size='1ms', win_size='1ms', sample='mean', sig_limit=0.05):
    @self.iter('average')
    def to_signif(case_raw_data):
        case_raw_data = sampling(case_raw_data, step_size, win_size, sample)
            
        data_with_subject = case_raw_data.mean(level=['subject','condition_group','channel'])
        
        check_availability(data_with_subject, 'condition_group','==2')

        stats_result = roll_on_levels_and_compare(data_with_subject, stats_methods.t_test, levels=['time','channel'], 
                          between='condition_group', in_group='subject',prograssbar=True)
  
        stats_result = stats_result.stack('time').unstack('channel').apply(lambda x:sum(x<sig_limit),axis=1).unstack('time')

        return stats_result

    signif_collection = to_signif()

    default_plot_params = dict(title='significant_channels_count',plot_type=['direct','heatmap'], x_len=12,
                                color=sns.cubehelix_palette(light=1, as_cmap=True), x_title='time', y_title='condition_group',cbar_title='Count')
    return structure.Analyzed_data('significant_channels_count', signif_collection, default_plot_params=default_plot_params)
    



# from ..default import *
# from ..condition import *
# from .. import draw
# from ..group_template import group_template


# def Latency(data,groups,window,sig_limit=0.1,func='max'):
#     def peak(x):
#         if func=='min':
#             index = signal.argrelmin(x, order=3)[0]
#             if(len(index)==0):
#                 return 0
#             extremeValue = min([x[i] for i in index])
#             individual_peaks = np.apply_along_axis(minP, 2, individual_win)
#         elif func=='max':
#             index = signal.argrelmax(x, order=3)[0]
#             if(len(index)==0):
#                 return 0
#             extremeValue = max([x[i] for i in index])
#         extremePoint = list(filter(lambda i:x[i]==extremeValue,index))[0]
#         return extremePoint
    
#     func1 = lambda sub: sub.data.mean(0)
#     func2 = lambda sub_1,sub_2: sub_1.data.mean(0)-sub_2.data.mean(0)
    
#     lines_group = group_template(data,groups,'ERP',func1,func2)
#     win = range((window[0]-epoch_time[0])*sr//1000,(window[1]-epoch_time[0])*sr//1000+1)

#     for subject_lines in lines_group: 
#         individual_win = np.array(subject_lines[2])[:,:,win]
#         individual_peaks = np.apply_along_axis(peak, 2, individual_win)
#         #deal with miss values
#         individual_peaks = np.array([individual_peaks[:,i] for i in range(individual_peaks.shape[1]) if min(individual_peaks[:,i])>0]).T
#         #t-test
#         T_result = scipy.stats.ttest_rel(individual_peaks[0],individual_peaks[1])
#         if T_result[1]<sig_limit:
#             print(subject_lines[0],subject_lines[1],window,[round(i,2) for i in individual_peaks.mean(1)],round(T_result[1],4))
#     #     sns.tsplot(individual_peaks.T,err_style="ci_bars",interpolate=False
    
# def Amplitude(data,groups,window,sig_limit=0.1,func=np.mean):
#     func1 = lambda sub: sub.data.mean(0)
#     func2 = lambda sub_1,sub_2: sub_1.data.mean(0)-sub_2.data.mean(0)
    
    
    
#     lines_group = group_template(data,groups,'ERP',func1,func2)
#     win = range((window[0]-epoch_time[0])*sr//1000,(window[1]-epoch_time[0])*sr//1000+1)

#     for subject_lines in lines_group: 
#         individual_win = np.array(subject_lines[2])[:,:,win]
#         individual_peaks = np.apply_along_axis(func, 2, individual_win)
#         #t-test
#         T_result = scipy.stats.ttest_rel(individual_peaks[0],individual_peaks[1])
#         if T_result[1]<sig_limit:
#             print(str(func).split(' ')[2][:-1],subject_lines[0][4:],window,subject_lines[1],
#                   ['%.2f' %i for i in individual_peaks.mean(1)],'%.4f' %T_result[1])
#     #     sns.tsplot(individual_peaks.T,err_style="ci_bars",interpolate=False



# coding:utf-8
# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name = 'easyEEG',
    version = '0.8.2.2',
    keywords = ['experiment', 'eeg', 'analysis', 'analyzer', 'neuroscience'],
    description = 'Concise, agile, and flexible EEG toolbox',
    author = 'Jinbiao Yang',
    author_email = 'ray306@gmail.com',
    long_description=long_description,
    license = 'GPL License',
    packages=find_packages(exclude=[".git"]),
    install_requires = ['numpy','pandas','scipy','matplotlib','statsmodels','seaborn','mne','permute','tqdm','ipdb'],

    include_package_data=True,

    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Programming Language :: Python :: 3',
    ],
)
